-------------------------------------------------------------------------------
--                               Lock Manager                                --
--                                                                           --
--  This is the root type for the lock manager                               --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/02/12 14:15:30 $
--  $RCSfile: lock_managers.adb,v $
--  $Revision: 1.9 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Ada.Tags; use Ada.Tags;
with Debug; use Debug;

package body Lock_Managers is

   function Is_Compatible (Set : Lock_Set_Type;
                           Lock : Lock_Ref;
                           Transaction : Transaction_Ref)
      return Boolean;

   function Is_Compatible (Set : Lock_Set_Type;
                           Lock : Lock_Ref;
                           Transaction : Transaction_Ref)
      return Boolean is
      Cursor : Lock_Node_Ref := Set.Head;
   begin
      while Cursor /= null loop
         if not Is_Ancestor_Or_Same (Transaction.all, Cursor.Transaction)
           and then not Locks.Is_Compatible (Lock.all, Cursor.Lock.all) then
            Put_Line ("Not compatible lock " & External_Tag (Lock.all'Tag));
            return False;
         end if;
         Cursor := Cursor.Next;
      end loop;
      Put_Line ("Compatible lock");
      return True;
   end Is_Compatible;

   procedure Insert (Set : in out Lock_Set_Type;
                     Lock : Lock_Ref;
                     Transaction : Transaction_Ref);

   procedure Insert (Set : in out Lock_Set_Type;
                     Lock : Lock_Ref;
                     Transaction : Transaction_Ref) is
      Temp : Lock_Node_Ref := Set.Head;
      Found : Boolean := False;
   begin
      Put_Line ("In insert lock");
      while Temp /= null loop
         if Temp.Transaction = Transaction and then
           Temp.Lock = Lock then
            --  If the lock is already in the list, we don't have to
            --  insert it
            Found := True;
            Put_Line ("Already inserted");
            exit;
         end if;
         Temp := Temp.Next;
      end loop;
      if not Found then
         Temp := new Lock_Node_Type;
         Temp.Lock := Lock;
         Temp.Transaction := Transaction;
         Temp.Next := Set.Head;
         Set.Head := Temp;
         Put_Line ("Lock inserted");
      end if;
   end Insert;

   procedure Free is new Ada.Unchecked_Deallocation
     (Lock_Node_Type, Lock_Node_Ref);

   procedure Delete (Set : in out Lock_Set_Type;
                     Transaction : Transaction_Ref);

   procedure Delete (Set : in out Lock_Set_Type;
                     Transaction : Transaction_Ref) is

      Temp    : Lock_Node_Ref := Set.Head;
      Prev    : Lock_Node_Ref := Set.Head;
      To_Free : Lock_Node_Ref;
   begin
      --  Run through all the nodes and delete all nodes
      --  that are related to the transaction
      while Temp /= null loop
         if Temp.Transaction = Transaction then
            if Temp = Set.Head then
               Set.Head := Temp.Next;
            else
               Prev.Next := Temp.Next;
            end if;

            --  The lock must not be freed, since it is static
            --  in the proxy
            Put_Line ("Deleted lock");
            To_Free := Temp;
            Temp := Temp.Next;
            Free (To_Free);
         else
            Prev := Temp;
            Temp := Temp.Next;
         end if;
      end loop;
   end Delete;

   procedure Pass_Up (Set : in out Lock_Set_Type;
                      Transaction : Transaction_Ref;
                      Parent : Transaction_Ref);

   procedure Pass_Up (Set : in out Lock_Set_Type;
                      Transaction : Transaction_Ref;
                      Parent : Transaction_Ref) is
      Temp, Prev  : Lock_Node_Ref := Set.Head;
      Parent_Temp : Lock_Node_Ref;
      To_Free : Lock_Node_Ref;
      Found : Boolean;
      Parent_Transaction : Transaction_Ref := Get_Parent (Transaction.all);
   begin
      --  Run through all the nodes
      while Temp /= null loop
         if Temp.Transaction = Transaction then

            --  Check if the parent already has it
            Found := False;
            Parent_Temp := Set.Head;
            while Parent_Temp /= null loop
               if Parent_Transaction = Parent_Temp.Transaction and then
                 Temp.Lock = Parent_Temp.Lock then
                  --  The parent already has this lock
                  Found := True;
                  exit;
               end if;
               Parent_Temp := Parent_Temp.Next;
            end loop;

            if Found then
               Put_Line ("Parent has lock already -> delete");
               if Temp = Set.Head then
                  Set.Head := Temp.Next;
               else
                  Prev.Next := Temp.Next;
               end if;
               To_Free := Temp;
               Temp := Temp.Next;
               Free (To_Free);
            else
               --  We found a lock that must be passed up
               Put_Line ("Passing lock to parent");
               Temp.Transaction := Parent;
               Temp := Temp.Next;
            end if;
         else
            Prev := Temp;
            Temp := Temp.Next;
         end if;
      end loop;
   end Pass_Up;

   protected body Protected_Lock_Manager_Type is

      entry Request_Lock_And_Mutex (Lock : Lock_Ref;
                                    Transaction  : Transaction_Ref)
         when True is
      begin
         Put_Line ("In request lock and mutex");
         --  If the task requesting the lock is part/or a child of
         --  the transactions that already hold locks,
         --  or if the operation commutes (the requested lock is
         --  compatible to all other held locks), then we can
         --  grant the new lock.
         if not Is_Compatible (My_Locks, Lock, Transaction) then
            --  We must wait
            requeue Waiting_For_Lock with abort;
         else
            --  We grant the lock, therefore we add it to the locks held list
            Insert (My_Locks, Lock, Transaction);
            --  We must now check for mutual exclusion
            requeue Waiting_For_Writer with abort;
         end if;
      end Request_Lock_And_Mutex;

      entry Waiting_For_Lock (Lock : Lock_Ref;
                              Transaction : Transaction_Ref)
         when To_Try > 0 is
      begin
         To_Try := To_Try - 1;
         requeue Request_Lock_And_Mutex with abort;
      end Waiting_For_Lock;

      entry Waiting_For_Writer (Lock : Lock_Ref;
                               Transaction : Transaction_Ref)
         when not Currently_Writing and Waiting_For_Readers'Count = 0 is
      begin
         if Is_Update (Lock.all) then
            --  The operation that will be called is an update operation
            if Currently_Reading > 0 then
               --  We already have a reader operations executing
               --  -> block the call
               requeue Waiting_For_Readers with abort;
            else
               --  Grant the mutex
               Currently_Writing := True;
            end if;
         else
            --  The operation that will be called is a read-only operation
            Currently_Reading := Currently_Reading + 1;
         end if;
      end Waiting_For_Writer;

      entry Waiting_For_Readers (Lock : Lock_Ref;
                                 Transaction : Transaction_Ref)
         when Currently_Reading = 0 is
      begin
         Currently_Writing := True;
      end Waiting_For_Readers;

      procedure Free_Mutex is
      begin
         if Currently_Writing then
            Currently_Writing := False;
         else
            Currently_Reading := Currently_Reading - 1;
         end if;
      end Free_Mutex;

      procedure Transaction_Commit (Transaction : Transaction_Ref) is
      begin
         if Is_Toplevel (Transaction.all) then
            --  We can safely delete the locks
            Delete (My_Locks, Transaction);
         else
            --  We must pass the locks up to our parent
            Pass_Up (My_Locks, Transaction, Get_Parent (Transaction.all));
         end if;
         To_Try := Waiting_For_Lock'Count;
      end Transaction_Commit;

      procedure Transaction_Abort (Transaction : Transaction_Ref) is
      begin
         --  We can safely delete all locks, since the transaction
         --  aborts. Locks that were inherited from the parent are still
         --  held by the parent
         Delete (My_Locks, Transaction);
         To_Try := Waiting_For_Lock'Count;
      end Transaction_Abort;

   end Protected_Lock_Manager_Type;

   procedure Request_Lock_And_Mutex (Lock_Manager : Lock_Manager_Type;
                                     Operation : Normal_Operation_Type'Class;
                                     Transaction : Transaction_Ref)
   is
      Lock : Lock_Ref := Get_Kind_Of_Lock (Operation);
   begin
      Lock_Manager.My_Lock_Manager.Request_Lock_And_Mutex (Lock, Transaction);
   end Request_Lock_And_Mutex;

   procedure Free_Mutex (Lock_Manager : Lock_Manager_Type) is
   begin
      Lock_Manager.My_Lock_Manager.Free_Mutex;
   end Free_Mutex;

   procedure Transaction_Commit (Lock_Manager : Lock_Manager_Type;
                                 Transaction : Transaction_Ref) is
   begin
      Lock_Manager.My_Lock_Manager.Transaction_Commit (Transaction);
   end Transaction_Commit;

   procedure Transaction_Abort (Lock_Manager : Lock_Manager_Type;
                                 Transaction : Transaction_Ref) is
   begin
      Lock_Manager.My_Lock_Manager.Transaction_Abort (Transaction);
   end Transaction_Abort;

   procedure Initialize (Manager : in out Lock_Manager_Type) is
   begin
      Manager.My_Lock_Manager := new Protected_Lock_Manager_Type;
   end Initialize;

   procedure Free is new Ada.Unchecked_Deallocation
     (Protected_Lock_Manager_Type, Protected_Lock_Manager_Ref);

   procedure Finalize (Manager : in out Lock_Manager_Type) is
   begin
      Free (Manager.My_Lock_Manager);
   end Finalize;


end Lock_Managers;
