-------------------------------------------------------------------------------
--                      Memory Objects.Concrete.Undobale                     --
--                                                                           --
--  This is the implementation of Undo_Redoable memory objects.              --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/24 07:06:52 $
--  $RCSfile: memory_objects-concrete-undo_redoable.adb,v $
--  $Revision: 1.10 $
--                                                                           --
-------------------------------------------------------------------------------

with Logging_Techniques; use Logging_Techniques;

with Debug; use Debug;
with Ada.Exceptions; use Ada.Exceptions;
with Ada.Unchecked_Deallocation;

package body Memory_Objects.Concrete.Undo_Redoable is

   use Undo_Lists;

   use Transaction_Lists;

   Local_Transaction_Not_Found : exception renames No_Such_Key;

   procedure Free is new Ada.Unchecked_Deallocation
     (Logging_Technique_Type'Class, Logging_Technique_Ref);

   function Create (Cache : Cache_Manager_Ref;
                    Logging_Factory : Logging_Factory_Ref)
      return Undo_Redoable_Object_Ref is
      Result : Undo_Redoable_Object_Ref := new Undo_Redoable_Object_Type;
   begin
      Result.Factory := Logging_Factory;
      Result.My_Cache_Manager := Cache;
      return Undo_Redoable_Object_Ref (Result);
   end Create;

   procedure Execute (Object      : in out Undo_Redoable_Object_Type;
                      Operation   : in out Normal_Operation_Type'Class;
                      Transaction : in Transaction_Ref) is
      Is_Physical : Boolean := Logging_Techniques.Is_Physical
        (Object.Factory.all);
   begin
      if Is_Update (Operation) and Is_Physical then
         --  we might need to update out physical undo-list
         declare
            Tmp : Element_Node_Ref;
         begin
            Tmp := Find (Object.Undo_List, Transaction);
            --  The logging is physical, we don't need to log again
         exception
            when Local_Transaction_Not_Found =>
               declare
                  New_Technique : Logging_Technique_Ref := Create
                    (Object.Factory.all, Operation,
                     Object.My_Saving_Operation, Object.My_Data_Object);
               begin
                  Insert_Element (Object.Undo_List, Transaction,
                                  New_Technique, True);
                  Put_Line ("Memory: Created new physical undo-list");
               end;
         end;
      end if;

      if Is_Update (Operation) then
         Insert_If_Not_Already (Object.Trans_List, Transaction);
      end if;

      --  now we really execute the operation
      Execute (Memory_Object_Type (Object), Operation, Transaction);
      Put_Line ("Memory: Executed operation for transaction" &
                Transaction_Text (Transaction.all) & " on object " &
                Storage_Params_To_String (Get_Params (Object)));

      if Is_Update (Operation) then
         if not Is_Physical then
            --  we must update our logical undo-list
            declare
               New_Technique : Logging_Technique_Ref;
            begin
               --  We must insert the operation into the already existing
               --  or not undo list
               New_Technique := Create
                 (Object.Factory.all, Operation,
                  Object.My_Saving_Operation, Object.My_Data_Object);
               Insert_Element (Object.Undo_List, Transaction,
                               New_Technique, True);
               Put_Line ("Memory: Inserted logical operation to Undo_List");

            end;
         else
            --  we must update our physical redo list
            declare
               Tmp : Element_Node_Ref;
            begin
               Tmp := Find (Object.Redo_List, Transaction);
               --  The redo logging is physical, we must replace the old value
               --  by the new one
               declare
                  Old_Technique : Logging_Technique_Ref := Tmp.Element;
               begin
                  --  create the new technique
                  Tmp.Element := Create
                    (Object.Factory.all, Operation,
                     Object.My_Saving_Operation, Object.My_Data_Object);
                  Free (Old_Technique);
               end;
               Put_Line ("Memory: Replaced Redo_List");
            exception
               when Local_Transaction_Not_Found =>
                  declare
                     New_Technique : Logging_Technique_Ref := Create
                       (Object.Factory.all, Operation,
                        Object.My_Saving_Operation, Object.My_Data_Object);
                  begin
                     Insert_Element (Object.Redo_List, Transaction,
                                     New_Technique, True);
                     Put_Line ("Memory: Created new Redo_List");
                  end;
            end;
         end if;
      end if;
   end Execute;

   function Get_Undo_List (Object : Undo_Redoable_Object_Type;
                           Transaction : Transaction_Ref) return
     Logging_Technique_Array_Type is
      Tmp : Element_Node_Ref;
      Undo_List : Element_Node_Ref;
      N : Natural := 0;
   begin
      Undo_List := Find (Object.Undo_List, Transaction);
      if Undo_List = null then
         raise Transaction_Not_Found;
      end if;
      Tmp := Undo_List;
      while Tmp /= null loop
         N := N + 1;
         Tmp := Tmp.Next;
      end loop;
      declare
         Result : Logging_Technique_Array_Type (1 .. N);
         I : Natural := 1;
      begin
         Tmp := Undo_List;
         while Tmp /= null loop
            Result (I) := Tmp.Element;
            Tmp := Tmp.Next;
            I := I + 1;
         end loop;
         return Result;
      end;
   exception
      when No_Such_Key =>
         raise Transaction_Not_Found;
   end Get_Undo_List;

   function Get_Redo_List (Object : Undo_Redoable_Object_Type;
                           Transaction : Transaction_Ref) return
     Logging_Technique_Array_Type is
      Tmp : Element_Node_Ref;
      Redo_List : Element_Node_Ref;
      Is_Physical : Boolean := Logging_Techniques.Is_Physical
        (Object.Factory.all);
      N : Natural := 0;
   begin
      Put_Line ("Memory: Get_Redo_List called for object " &
                Storage_Params_To_String (Get_Params (Object)));
      if Is_Physical then
         Redo_List := Find (Object.Redo_List, Transaction);
      else
         --  the logical undo list is also the redo list!
         Redo_List := Find (Object.Undo_List, Transaction);
      end if;

      if Redo_List = null then
         raise Transaction_Not_Found;
      end if;
      Tmp := Redo_List;
      while Tmp /= null loop
         N := N + 1;
         Tmp := Tmp.Next;
      end loop;
      declare
         Result : Logging_Technique_Array_Type (1 .. N);
         I : Natural := 1;
      begin
         Tmp := Redo_List;
         while Tmp /= null loop
            Result (I) := Tmp.Element;
            Tmp := Tmp.Next;
            I := I + 1;
         end loop;
         return Result;
      end;
   exception
      when No_Such_Key =>
         raise Transaction_Not_Found;
   end Get_Redo_List;

   procedure Real_Abort (Object : in out Undo_Redoable_Object_Type;
                         Transaction : in Transaction_Ref) is
      Tmp : Element_Node_Ref;
   begin
      Tmp := Find (Object.Undo_List, Transaction);

      if Is_Physical (Object.Factory.all) then
         --  we must undo the transaction, and then delete the undo list
         Logging_Techniques.Undo (Tmp.Element.all, Self (Object));
         Put_Line ("Previous version restored using physical undo");
         Delete (Object.Undo_List, Transaction);
         --  we must also delete the physical redo list
         Delete (Object.Redo_List, Transaction);
      else
         --  logical logging
         --  we must undo all operations and
         --  we must delete the undo list
         while Tmp /= null loop
            Undo (Tmp.Element.all, Self (Object));
            Put_Line ("Previous version restored using logical undo");
            Tmp := Tmp.Next;
         end loop;
         Delete (Object.Undo_List, Transaction);
         --  we don't have to delete the redo-list, since the
         --  logical undo-list is also the redo list
      end if;
      Delete_Element (Object.Trans_List, Transaction);
   exception
      when Local_Transaction_Not_Found =>
         --  The transaction has not accessed any objects
         null;
      when Transaction_Lists.No_Such_Element =>
         null;
   end Real_Abort;

   procedure Real_Commit (Object : in out Undo_Redoable_Object_Type;
                          Transaction : in Transaction_Ref) is
      Child_Undo_List : Element_Node_Ref;
   begin
      Put_Line ("Memory: Called Real_Commit for object " &
                Storage_Params_To_String (Get_Params (Object)));
      Child_Undo_List := Find (Object.Undo_List, Transaction);

      if not Is_Toplevel (Transaction.all) then
         --  we are a nested transaction, we must now move our
         --  undo list to the parent, if necessary
         Put_Line ("Memory: Nested transaction");
         declare
            Parent_Undo_List : Element_Node_Ref;
            Parent : Transaction_Ref := Get_Parent (Transaction.all);
         begin
            Parent_Undo_List := Find (Object.Undo_List, Parent);
            Put_Line ("Memory_Object: Found parent undo-list");
            if not Is_Physical (Object.Factory.all) then
               --  We must prepend the child undo list to the parent undo-list
               Put_Line ("Memory: Merging logical undo-list with parent");
               Merge_Lists (Object.Undo_List, Parent, Transaction);
            else
               --  we must do nothing for the physical undo list
               --  but we must replace the physical redo-image, if there
               --  is already one
               Put_Line ("Memory: Replaced physical redo-list");
               Delete (Object.Redo_List, Parent);
               Replace_Key (Object.Redo_List, Transaction, Parent);
            end if;

         exception
            when Local_Transaction_Not_Found =>
               --  The child undo-list and redo-list must be given to parent
               Put_Line ("Memory: Moved undo-list to parent");
               Replace_Key (Object.Undo_List, Transaction, Parent);
               if Is_Physical (Object.Factory.all) then
                  Put_Line ("Memory: Moved physical redo-list to parent");
                  Replace_Key (Object.Redo_List, Transaction, Parent);
               end if;
            when E : others =>
               Put_Line ("Memory: Exception " & Exception_Name (E));
               raise;
         end;
         --  we must also hand over our transaction list
         begin
            Delete_Element (Object.Trans_List, Transaction);
            Insert_If_Not_Already (Object.Trans_List,
                                   Get_Parent (Transaction.all));
         exception
            when Transaction_Lists.No_Such_Element =>
               null;
         end;
      else
         --  This is a toplevel transaction
         --  We can now forget the undo operations
         Put_Line ("Memory: Toplevel transaction");
         Delete (Object.Undo_List, Transaction);
         Delete_Element (Object.Trans_List, Transaction);
      end if;
   exception
      when Local_Transaction_Not_Found =>
         --  This object has been accessed read-only
         null;
      when E : others =>
         Put_Line ("Memory: Exception " & Exception_Name (E));
         raise;
   end Real_Commit;

   procedure Finalize (Object : in out Undo_Redoable_Object_Type) is
   begin
      null;
   end Finalize;

   function OK_To_Propagate (Object : Undo_Redoable_Object_Type)
      return Boolean is
   begin
      return Count_Elements (Object.Trans_List) = 0;
   end OK_To_Propagate;

end Memory_Objects.Concrete.Undo_Redoable;


