-------------------------------------------------------------------------------
--                      Memory Objects.Concrete.Versioned                    --
--                                                                           --
--  This is the implementation of versioned memory objects.                  --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/08/18 11:48:40 $
--  $RCSfile: memory_objects-concrete-versioned.adb,v $
--  $Revision: 1.1 $
--                                                                           --
-------------------------------------------------------------------------------

with Debug; use Debug;
with Ada.Unchecked_Deallocation;

package body Memory_Objects.Concrete.Versioned is

   procedure Free_Stream is new Ada.Unchecked_Deallocation
     (Stream_Type'Class, Stream_Ref);

   procedure Free is new Ada.Unchecked_Deallocation
     (Node_Type, Node_Ref);

   procedure Free_Memory is new Ada.Unchecked_Deallocation
     (Memory_Storage_Type'Class, Memory_Storage_Ref);

   function Create (Cache : Cache_Manager_Ref) return Versioned_Object_Ref is
      Result : Versioned_Object_Ref := new Versioned_Object_Type;
   begin
      Result.My_Cache_Manager := Cache;
      return Versioned_Object_Ref (Result);
   end Create;

   procedure Execute (Object      : in out Versioned_Object_Type;
                      Operation   : in out Normal_Operation_Type'Class;
                      Transaction : in Transaction_Ref) is
   begin
      if Is_Update (Operation) then
         --  If this is an update operation, we might need to create
         --  a new version of the objects state
         Check_And_Then_Insert (Object, Transaction);
      end if;
      Execute (Memory_Object_Type (Object), Operation, Transaction);
   end Execute;

   procedure Check_And_Then_Insert
     (Object      : in out Versioned_Object_Type;
      Transaction : in Transaction_Ref) is
      Tmp : Node_Ref := Object.Versionlist;
      Found : Boolean := False;
   begin
      while Tmp /= null loop
         if Tmp.Transaction = Transaction then
            Found := True;
            exit;
         end if;
         Tmp := Tmp.Next;
      end loop;
      if not Found then
         Tmp := new Node_Type;
         Tmp.Transaction := Transaction;
         Tmp.Next := Object.Versionlist;
         Object.Versionlist := Tmp;
         Tmp.State := new Memory_Storage_Type;
         declare
            Temp_Stream : Stream_Ref := new Stream_Type (Tmp.State);
         begin
            Save (Object.My_Saving_Operation.all,
                  Object.My_Data_Object,
                  Abstract_Stream_Ref (Temp_Stream));
            Free_Stream (Temp_Stream);
         end;
      end if;
   end Check_And_Then_Insert;

   procedure Real_Discard (Object : in out Versioned_Object_Type;
                           Transaction : in Transaction_Ref) is
      Tmp : Node_Ref := Object.Versionlist;
      Prev : Node_Ref := Object.Versionlist;
   begin
      while Tmp /= null loop
         if Tmp.Transaction = Transaction then
            declare
               Temp_Stream : Stream_Ref :=
                new Stream_Type (Tmp.State);
            begin
               Load (Object.My_Loading_Operation.all,
                     Object.My_Data_Object,
                     Abstract_Stream_Ref (Temp_Stream));
               Put_Line ("Previous version restored");
               Free_Stream (Temp_Stream);
               if Tmp = Object.Versionlist then
                  --  It's the first node, so let's remove it
                  Object.Versionlist := Tmp.Next;
               else
                  Prev.Next := Tmp.Next;
               end if;
               Free_Memory (Tmp.State);
               Free (Tmp);
            end;
            return;
         end if;
         Prev := Tmp;
         Tmp := Tmp.Next;
      end loop;
   end Real_Discard;

end Memory_Objects.Concrete.Versioned;


