-------------------------------------------------------------------------------
--                           Memory Objects.Concrete                         --
--                                                                           --
--  This is the concrete root type for the memory objects.                   --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/24 07:06:53 $
--  $RCSfile: memory_objects-concrete.adb,v $
--  $Revision: 1.10 $
--                                                                           --
-------------------------------------------------------------------------------

with Storages; use Storages;
with Storages.Non_Volatile; use Storages.Non_Volatile;

with Streams; use Streams;
with Buffers.Unbounded; use Buffers.Unbounded;

with Ada.Unchecked_Deallocation;

with Debug; use Debug;

package body Memory_Objects.Concrete is

   procedure Free_Stream is new Ada.Unchecked_Deallocation
     (Stream_Type'Class, Stream_Ref);

   function Create (Cache : Cache_Manager_Ref) return Memory_Object_Ref is
      Result : Concrete_Memory_Object_Ref := new Memory_Object_Type;
   begin
      Result.My_Cache_Manager := Cache;
      return Memory_Object_Ref (Result);
   end Create;

   procedure Create (Object   : in out Memory_Object_Type;
                     Params   : Non_Volatile_Params_Type'Class;
                     Creation : Creation_Operation_Type'Class;
                     Loading  : Loading_Operation_Type'Class;
                     Saving   : Saving_Operation_Type'Class;
                     Deletion : Deletion_Operation_Type'Class) is
   begin
      --  First try and create a real storage
      --  This can fail e.g. if the storage already exists
      Object.My_Storage := Non_Volatile_Storage_Ref
        (Create_Storage (Params));

      --  Now we call the Init procedure
      Init (Memory_Object_Type'Class (Object),
            Params, Loading, Saving, Deletion);

      --  Call the creation operation
      Object.My_Data_Object := Create (Creation);

      declare
         Temp_Stream : Stream_Ref :=
          new Stream_Type (Object.My_Storage);
      begin
         Natural'Write (Temp_Stream, Initial_State);
         Save (Saving,
               Object.My_Data_Object,
               Abstract_Stream_Ref (Temp_Stream));
         Free_Stream (Temp_Stream);
      end;

      --  Close the storage
      Close (Object.My_Storage.all);
      Put_Line ("Created");
   exception
      when Storage_Existant =>
         raise Object_Exists;
   end Create;

   procedure Restore (Object   : in out Memory_Object_Type;
                      Params   : Non_Volatile_Params_Type'Class;
                      Loading  : Loading_Operation_Type'Class;
                      Saving   : Saving_Operation_Type'Class;
                      Deletion : Deletion_Operation_Type'Class) is
   begin
      --  First try and open the storage
      --  This might fail e.g. if the storage does not exist
      Object.My_Storage := Open_Storage (Params);

      --  Now we call the Init procedure
      Init (Memory_Object_Type'Class (Object),
            Params, Loading, Saving, Deletion);

      --  Now we will try and read the state from the storage
      declare
         Temp_Stream : Stream_Ref :=
             new Stream_Type (Object.My_Storage);
         State_Identifier : Natural;
      begin
         --  Call the restore operation
         Natural'Read (Temp_Stream, State_Identifier);
         Object.My_State_Identifier := State_Identifier;
         Create_And_Load (Object.My_Loading_Operation.all,
                          Object.My_Data_Object,
                          Abstract_Stream_Ref (Temp_Stream));
         Free_Stream (Temp_Stream);
      end;

      --  Close the storage
      Close (Object.My_Storage.all);
      Put_Line ("Restored");
   exception
      when Storage_Not_Existant =>
         raise Object_Non_Existant;
      when Data_Not_Available =>
         Close (Object.My_Storage.all);
         raise Object_Corrupt;
   end Restore;

   procedure Execute (Object      : in out Memory_Object_Type;
                      Operation   : in out Normal_Operation_Type'Class;
                      Transaction : in Transaction_Ref) is
   begin
      Do_Operation (Operation, Object.My_Data_Object);
   end Execute;

   procedure Execute_Undo (Object      : in out Memory_Object_Type;
                           Operation   : in out Normal_Operation_Type'Class;
                           Transaction : in Transaction_Ref) is
   begin
      Undo_Operation (Operation, Object.My_Data_Object);
   end Execute_Undo;

   procedure Execute_Load (Object : in out Memory_Object_Type;
                           Stream : in Abstract_Stream_Ref) is
   begin
      Load (Object.My_Loading_Operation.all, Object.My_Data_Object, Stream);
   end Execute_Load;

   procedure Free_Stream is new Ada.Unchecked_Deallocation
     (Buffered_Stream_Type'Class, Buffered_Stream_Ref);

   procedure Free_Buffer is new Ada.Unchecked_Deallocation
     (Unbounded_Buffer_Type'Class, Unbounded_Buffer_Ref);

   protected body Mutex_Type is

      entry Propagate (Object : in out Memory_Object_Type)
      when not Pinned is
      begin
         if OK_To_Propagate (Memory_Object_Type'Class (Object)) then
            Put_Line ("Memory_Object : OK to propagate "
                      & Storage_Params_To_String (Object.My_Params.all));
            declare
               Temp_Buffer : Unbounded_Buffer_Ref :=
                new Unbounded_Buffer_Type (Object.My_Storage);
               Temp_Stream : Buffered_Stream_Ref :=
                new Buffered_Stream_Type (Temp_Buffer);
            begin
               Object.My_State_Identifier := Object.My_State_Identifier + 1;
               Open (Object.My_Storage.all, Out_Mode);
               Natural'Write (Temp_Stream, Object.My_State_Identifier);
               Save (Object.My_Saving_Operation.all,
                     Object.My_Data_Object, Abstract_Stream_Ref (Temp_Stream));
               Flush (Temp_Stream.all);
               Close (Object.My_Storage.all);
               Free_Stream (Temp_Stream);
               Free_Buffer (Temp_Buffer);
            end;
         else
            Put_Line ("Memory_Object : Propagation of " &
                      Storage_Params_To_String (Object.My_Params.all) &
                      " delayed!", Important);
         end if;
      end Propagate;

      procedure Abort_Transaction (Object : in out Memory_Object_Type;
                                   Transaction : in Transaction_Ref) is
      begin
         Real_Abort (Memory_Object_Type'Class (Object), Transaction);
      end Abort_Transaction;

      procedure Commit_Transaction (Object : in out Memory_Object_Type;
                                    Transaction : in Transaction_Ref) is
      begin
         Real_Commit (Memory_Object_Type'Class (Object), Transaction);
      end Commit_Transaction;

      procedure Pin is
      begin
         Pinned := True;
      end Pin;

      procedure Un_Pin is
      begin
         Pinned := False;
      end Un_Pin;

   end Mutex_Type;

   procedure Propagate (Object : in out Memory_Object_Type) is
   begin
      Object.My_Mutex.Propagate (Object);
   end Propagate;

   procedure Abort_Transaction (Object : in out Memory_Object_Type;
                                Transaction : in Transaction_Ref) is
   begin
      Object.My_Mutex.Abort_Transaction (Object, Transaction);
   end Abort_Transaction;

   procedure Commit_Transaction (Object : in out Memory_Object_Type;
                                Transaction : in Transaction_Ref) is
   begin
      Object.My_Mutex.Commit_Transaction (Object, Transaction);
   end Commit_Transaction;

   procedure Pin (Object : in out Memory_Object_Type) is
   begin
      Object.My_Mutex.Pin;
   end Pin;

   procedure Un_Pin (Object : in out Memory_Object_Type) is
   begin
      Object.My_Mutex.Un_Pin;
   end Un_Pin;

   function Get_Lock_Manager (Object : Memory_Object_Type)
      return Lock_Manager_Ref is
   begin
      return Object.My_Lock_Manager;
   end Get_Lock_Manager;

   function Self (M : in Memory_Object_Type) return Memory_Object_Ref is
   begin
      return M.Self.Obj.all'Unchecked_Access;
   end Self;

   procedure Init (Object   : in out Memory_Object_Type;
                   Params   : Non_Volatile_Params_Type'Class;
                   Loading  : Loading_Operation_Type'Class;
                   Saving   : Saving_Operation_Type'Class;
                   Deletion : Deletion_Operation_Type'Class) is
   begin
      Object.My_Lock_Manager := new Lock_Manager_Type;
      Object.My_Loading_Operation := new
         Loading_Operation_Type'Class' (Loading);
      Object.My_Saving_Operation := new
         Saving_Operation_Type'Class' (Saving);
      Object.My_Deletion_Operation := new
        Deletion_Operation_Type'Class' (Deletion);
      Object.My_Params := new Non_Volatile_Params_Type'Class' (Params);
      Put_Line ("Initialized");
   end Init;

   function Get_Params (Object : Memory_Object_Type) return
     Non_Volatile_Params_Type'Class is
   begin
      return Object.My_Params.all;
   end Get_Params;

   function Get_State_Identifier (Object : Memory_Object_Type)
     return State_Identifier_Type is
   begin
      return Object.My_State_Identifier;
   end Get_State_Identifier;

   procedure Real_Abort (Object : in out Memory_Object_Type;
                         Transaction : in Transaction_Ref) is
   begin
      Open (Object.My_Storage.all);

      declare
         Temp_Stream : Stream_Ref :=
          new Stream_Type (Object.My_Storage);
      begin
         --  Call the restore operation
         Load
           (Object.My_Loading_Operation.all,
            Object.My_Data_Object,
            Abstract_Stream_Ref (Temp_Stream));
         Free_Stream (Temp_Stream);
      end;
      Close (Object.My_Storage.all);
      Put_Line ("Discarded");
   end Real_Abort;

   procedure Real_Commit (Object : in out Memory_Object_Type;
                          Transaction : in Transaction_Ref) is
   begin
      null;
   end Real_Commit;

   function OK_To_Propagate (Object : Memory_Object_Type) return Boolean is
   begin
      return True;
   end OK_To_Propagate;

end Memory_Objects.Concrete;
