-------------------------------------------------------------------------------
--                           Memory Objects.Concrete                         --
--                                                                           --
--  This is the concrete root type for the memory objects.                   --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/24 07:06:53 $
--  $RCSfile: memory_objects-concrete.ads,v $
--  $Revision: 1.14 $
--                                                                           --
-------------------------------------------------------------------------------

with Streams; use Streams;

with Lock_Managers; use Lock_Managers;
with Cache_Managers; use Cache_Managers;
with Storages.Non_Volatile; use Storages.Non_Volatile;
with Storage_Params.Non_Volatile; use Storage_Params.Non_Volatile;
with Operations.Creation; use Operations.Creation;
with Operations.Normal; use Operations.Normal;
with Operations.Loading; use Operations.Loading;
with Operations.Saving; use Operations.Saving;
with Operations.Deletion; use Operations.Deletion;
with Transactions; use Transactions;

package Memory_Objects.Concrete is

   type Memory_Object_Type is new Abstract_Memory_Object_Type with private;
   type Concrete_Memory_Object_Ref is access all Memory_Object_Type'Class;

   function Create (Cache : Cache_Manager_Ref) return Memory_Object_Ref;
   --  Allocates a memory object

   procedure Create (Object   : in out Memory_Object_Type;
                     Params   : Non_Volatile_Params_Type'Class;
                     Creation : Creation_Operation_Type'Class;
                     Loading  : Loading_Operation_Type'Class;
                     Saving   : Saving_Operation_Type'Class;
                     Deletion : Deletion_Operation_Type'Class);

   procedure Restore (Object   : in out Memory_Object_Type;
                      Params   : Non_Volatile_Params_Type'Class;
                      Loading  : Loading_Operation_Type'Class;
                      Saving   : Saving_Operation_Type'Class;
                      Deletion : Deletion_Operation_Type'Class);

   procedure Execute (Object      : in out Memory_Object_Type;
                      Operation   : in out Normal_Operation_Type'Class;
                      Transaction : in Transaction_Ref);

   procedure Execute_Undo (Object      : in out Memory_Object_Type;
                           Operation   : in out Normal_Operation_Type'Class;
                           Transaction : in Transaction_Ref);

   procedure Execute_Load (Object : in out Memory_Object_Type;
                           Stream : in Abstract_Stream_Ref);

   procedure Propagate (Object : in out Memory_Object_Type);

   procedure Abort_Transaction (Object : in out Memory_Object_Type;
                                Transaction : in Transaction_Ref);

   procedure Commit_Transaction (Object : in out Memory_Object_Type;
                                 Transaction : in Transaction_Ref);

   procedure Pin (Object : in out Memory_Object_Type);

   procedure Un_Pin (Object : in out Memory_Object_Type);

   function Get_Lock_Manager (Object : Memory_Object_Type)
      return Lock_Manager_Ref;

   subtype State_Identifier_Type is Natural;

   Initial_State : constant Natural := 0;

   function Get_State_Identifier (Object : Memory_Object_Type)
     return State_Identifier_Type;

   Object_Exists, Object_Non_Existant, Object_Corrupt : exception;

   function Get_Params (Object : Memory_Object_Type) return
     Non_Volatile_Params_Type'Class;

private

   protected type Mutex_Type is

      entry Propagate (Object : in out Memory_Object_Type);

      procedure Abort_Transaction (Object : in out Memory_Object_Type;
                                   Transaction : in Transaction_Ref);

      procedure Commit_Transaction (Object : in out Memory_Object_Type;
                                    Transaction : in Transaction_Ref);

      procedure Pin;

      procedure Un_Pin;

   private

      Pinned : Boolean := False;

   end Mutex_Type;

   type Selfer (Obj : access Memory_Object_Type'Class) is
      limited null record;

   type Memory_Object_Type is new Abstract_Memory_Object_Type with record
      Self  : Selfer (Memory_Object_Type'Access);
      My_Data_Object : Data_Ref := null;
      My_Lock_Manager : Lock_Manager_Ref := null;
      My_Mutex : Mutex_Type;
      My_Cache_Manager : Cache_Manager_Ref := null;
      My_Loading_Operation : Loading_Operation_Ref := null;
      My_Saving_Operation : Saving_Operation_Ref := null;
      My_Deletion_Operation : Deletion_Operation_Ref := null;
      My_Params : Non_Volatile_Params_Ref := null;
      My_Storage : Non_Volatile_Storage_Ref := null;
      My_State_Identifier : Natural := Initial_State;
   end record;

   function Self (M : in Memory_Object_Type) return Memory_Object_Ref;

   procedure Init (Object   : in out Memory_Object_Type;
                   Params   : Non_Volatile_Params_Type'Class;
                   Loading  : Loading_Operation_Type'Class;
                   Saving   : Saving_Operation_Type'Class;
                   Deletion : Deletion_Operation_Type'Class);
   --  Initialize the components of the memory object

   procedure Real_Abort (Object : in out Memory_Object_Type;
                         Transaction : in Transaction_Ref);
   --  This is the operation that really discards the state of the
   --  data object. It is called from within the protected mutex.

   procedure Real_Commit (Object : in out Memory_Object_Type;
                          Transaction : in Transaction_Ref);
   --  This is the operation that really commits the state of the
   --  data object. It is called from within the protected mutex.

   function OK_To_Propagate (Object : Memory_Object_Type) return Boolean;
   --  This function must return true, if the object can currently be
   --  propagated to the storage device

end Memory_Objects.Concrete;
