
with Ada.Unchecked_Deallocation;

package body Persistent_Objects.Directories is

   procedure Free is new Ada.Unchecked_Deallocation (Object_Params_Node_Type,
                                                     Object_Params_Node_Ref);

   procedure Free is new Ada.Unchecked_Deallocation
     (Non_Volatile_Params_Type'Class,
      Non_Volatile_Params_Ref);

   function Create_Object (Storage_Params : in Non_Volatile_Params_Type'Class)
      return Directory_Ref is
      Result : Directory_Ref := new Directory_Type;
   begin
      Initialize_Object (Persistent_Object_Ref (Result), Storage_Params);
      return Result;
   end Create_Object;

   procedure Insert (Directory : in out Directory_Type;
                     Params    : in Non_Volatile_Params_Type'Class) is
      New_Node : Object_Params_Node_Ref := new Object_Params_Node_Type;
   begin
      New_Node.Params := new Non_Volatile_Params_Type'Class' (Params);
      New_Node.Next := Directory.Params_List;
      Directory.Params_List := New_Node;
      Directory.Real_Number := Directory.Real_Number + 1;
   end Insert;

   procedure Validate (Directory : in out Directory_Type;
                       Params    : in Non_Volatile_Params_Type'Class) is
      Temp : Object_Params_Node_Ref := Directory.Params_List;
   begin
      while Temp /= null and then Temp.Params.all /= Params loop
         Temp := Temp.Next;
      end loop;
      if Temp = null then
         raise Params_Not_Found;
      end if;
      Temp.Valid := True;
      Directory.Number := Directory.Number + 1;
   end Validate;

   procedure Check_Existence (Directory : in Directory_Type;
                              Params    : in Non_Volatile_Params_Type'Class;
                              Valid     : out Boolean;
                              Exists    : out Boolean) is
      Temp : Object_Params_Node_Ref := Directory.Params_List;
   begin
      while Temp /= null and then Temp.Params.all /= Params loop
         Temp := Temp.Next;
      end loop;
      if Temp = null then
         Valid := False;
         Exists := False;
      else
         Exists := True;
         Valid := Temp.Valid;
      end if;
   end Check_Existence;

   procedure Delete (Directory : in out Directory_Type;
                     Params    : in Non_Volatile_Params_Type'Class) is
      Temp : Object_Params_Node_Ref := Directory.Params_List;
      Old_Temp : Object_Params_Node_Ref;
   begin
      if Temp /= null and then Temp.Params.all = Params then
         Directory.Params_List := Temp.Next;
         if Temp.Valid then
            Directory.Number := Directory.Number - 1;
         end if;
         Directory.Real_Number := Directory.Real_Number - 1;
         Free (Temp.Params);
         Free (Temp);
      else
         while Temp /= null and then Temp.Params.all /= Params loop
            Old_Temp := Temp;
            Temp := Temp.Next;
         end loop;
         if Temp = null then
            raise Params_Not_Found;
         else
            Old_Temp.Next := Temp.Next;
            if Temp.Valid then
               Directory.Number := Directory.Number - 1;
            end if;
            Directory.Real_Number := Directory.Real_Number - 1;
            Free (Temp.Params);
            Free (Temp);
         end if;
      end if;
   end Delete;

   function Number_Of_Real_Objects (Directory : in Directory_Type)
                               return Natural is
   begin
      return Directory.Real_Number;
   end Number_Of_Real_Objects;

   function Number_Of_Valid_Objects (Directory : in Directory_Type)
                                     return Natural is
   begin
      return Directory.Number;
   end Number_Of_Valid_Objects;

   function Valid_Get (Directory : Directory_Type;
                       Number    : Natural)
                       return Non_Volatile_Params_Type'Class is
   begin
      if Directory.Number < Number then
         raise Params_Not_Found;
      else
         declare
            Temp : Object_Params_Node_Ref := Directory.Params_List;
            I    : Natural := Number;
         begin
            while not Temp.Valid loop
               Temp := Temp.Next;
            end loop;

            while I > 1 loop
               Temp := Temp.Next;
               if Temp.Valid then
                  I := I - 1;
               end if;
            end loop;
            return Temp.Params.all;
         end;
      end if;
   end Valid_Get;

   function Real_Get (Directory : Directory_Type;
                      Number    : Natural)
                      return Non_Volatile_Params_Type'Class is
   begin
      if Directory.Real_Number < Number then
         raise Params_Not_Found;
      else
         declare
            Temp : Object_Params_Node_Ref := Directory.Params_List;
            I    : Natural := Number;
         begin
            while I > 1 loop
               Temp := Temp.Next;
            end loop;
            return Temp.Params.all;
         end;
      end if;
   end Real_Get;

   procedure My_Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                       Item   : Directory_Type) is
      Temp : Object_Params_Node_Ref;
   begin
      Natural'Write (Stream, Item.Number);
      Natural'Write (Stream, Item.Real_Number);
      if Item.Real_Number /= 0 then
         Temp := Item.Params_List;
         while Temp /= null loop
            Non_Volatile_Params_Type'Class'Output (Stream, Temp.Params.all);
            Boolean'Write (Stream, Temp.Valid);
            Temp := Temp.Next;
         end loop;
      end if;
   end My_Write;

   procedure My_Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                      Item   : out Directory_Type) is
      Temp : Object_Params_Node_Ref;
      Old_Temp : Object_Params_Node_Ref := null;
   begin
      Natural'Read (Stream, Item.Number);
      Natural'Read (Stream, Item.Real_Number);
      for I in 1 .. Item.Real_Number loop
         Temp := new Object_Params_Node_Type;
         Temp.Params := new Non_Volatile_Params_Type'Class'
           (Non_Volatile_Params_Type'Class'Input (Stream));
         Boolean'Read (Stream, Temp.Valid);
         Temp.Next := Old_Temp;
         Old_Temp := Temp;
      end loop;
      Item.Params_List := Temp;
   end My_Read;

   procedure Deallocate_Object (Directory : in out Directory_Type) is
      Temp : Object_Params_Node_Ref;
      Temp_Free : Object_Params_Node_Ref;
   begin
      if Directory.Number /= 0 then
         Temp := Directory.Params_List;
         while Temp /= null loop
            Temp_Free := Temp;
            Temp := Temp.Next;
            Free (Temp_Free);
         end loop;
      end if;
   end Deallocate_Object;

end Persistent_Objects.Directories;
