with Ada.Unchecked_Deallocation;
with Buffers.Unbounded_Buffers;
use Buffers.Unbounded_Buffers;
with Storages; use Storages;
with Ada.IO_Exceptions; use Ada.IO_Exceptions;
with Streams; use Streams;

package body Persistent_Objects is

   procedure Free_Stream is new Ada.Unchecked_Deallocation
     (Buffered_Stream_Type'Class, Buffered_Stream_Ref);

   procedure Free_Buffer is new Ada.Unchecked_Deallocation
     (Unbounded_Buffer_Type'Class, Unbounded_Buffer_Ref);

   procedure Save_Object (Object : in out Persistent_Object_Type) is
      Temp_Buffer : Unbounded_Buffer_Ref :=
         new Unbounded_Buffer_Type (Object.Storage);
      Temp_Stream : Buffered_Stream_Ref :=
         new Buffered_Stream_Type (Temp_Buffer);
   begin
      Open (Object.Storage.all, Out_Mode);
      Persistent_Object_Type'Class'Output
        (Temp_Stream, Persistent_Object_Type'Class (Object));
      Flush (Temp_Stream.all);
      Close (Object.Storage.all);
      Free_Stream (Temp_Stream);
      Free_Buffer (Temp_Buffer);
   end Save_Object;

   procedure Free is new Ada.Unchecked_Deallocation
     (Stream_Type'Class, Stream_Ref);

   function Restore_Object (Storage_Params : in Non_Volatile_Params_Type'Class)
      return Persistent_Object_Ref is
      Temp_Storage_Ref : Non_Volatile_Storage_Ref;
   begin
      Temp_Storage_Ref :=
        Non_Volatile_Storage_Ref (Open_Storage (Storage_Params));
      declare
         Temp_Stream : Stream_Ref :=
            new Stream_Type (Non_Volatile_Storage_Ref (Temp_Storage_Ref));
         Temp_Object : Persistent_Object_Type'Class :=
           Persistent_Object_Type'Class'Input (Temp_Stream);
         Result : Persistent_Object_Ref :=
            new Persistent_Object_Type'Class' (Temp_Object);
      begin
         Free (Temp_Stream);
         Result.Storage := Temp_Storage_Ref;
         Result.Params := new Non_Volatile_Params_Type'Class'
           (Storage_Params);
         Close (Result.Storage.all);
         return Result;
      end;
   exception
      when Storage_Not_Existant =>
         raise Object_Non_Existant;
      when Data_Not_Available | End_Error =>
         Close (Temp_Storage_Ref.all);
         raise Object_Corrupt;
   end Restore_Object;

   procedure Initialize_Object
      (Object : in out Persistent_Object_Ref;
       Storage_Params : in Non_Volatile_Params_Type'Class) is
   begin
      Object.Storage := Non_Volatile_Storage_Ref
        (Create_Storage (Storage_Params));
      Object.Params := new Non_Volatile_Params_Type'Class'
        (Storage_Params);
      Close (Object.Storage.all);
   end Initialize_Object;

   procedure Free is new Ada.Unchecked_Deallocation
     (Non_Volatile_Storage_Type'Class, Non_Volatile_Storage_Ref);

   procedure Free is new Ada.Unchecked_Deallocation
     (Non_Volatile_Params_Type'Class, Non_Volatile_Params_Ref);

   procedure Finalize (S : in out Persistent_Object_Type) is
   begin
      if S.Storage /= null then
         Free (S.Storage);
      end if;
      if S.Params /= null then
         Free (S.Params);
      end if;
   end Finalize;

   pragma Warnings (Off);
   procedure My_Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                       Item   : in Persistent_Object_Type) is
   begin
      null;
   end My_Write;
   pragma Warnings (On);

   pragma Warnings (Off);
   procedure My_Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                      Item   : out Persistent_Object_Type) is
   begin
      null;
   end My_Read;
   pragma Warnings (On);

end Persistent_Objects;
