-------------------------------------------------------------------------------
--                                    Logs                                   --
--                                                                           --
--  This package declares the log type and its operations                    --
--                                                                           --
--  $Author: xcaron $
--  $Date: 2000/11/30 12:25:24 $
--  $RCSfile: logs.adb,v $
--  $Revision: 1.7 $
--                                                                           --
-------------------------------------------------------------------------------

--  with Log_Informations; use Log_Informations;
with Storages; use Storages;
with Storages.Non_Volatile; use Storages.Non_Volatile;
with Storages.Non_Volatile.Stable; use Storages.Non_Volatile.Stable;

with Streams; use Streams;
with Buffers.Unbounded; use Buffers.Unbounded;

with Logging_Techniques; use Logging_Techniques;

with Log_Information; use Log_Information;
with Log_Information.Transaction; use Log_Information.Transaction;
with Log_Information.Object.Undo; use Log_Information.Object.Undo;
with Log_Information.Object.Redo; use Log_Information.Object.Redo;
with Log_Information.Object.Create; use Log_Information.Object.Create;

with Memory_Objects.Concrete.Undo_Redoable;
use Memory_Objects.Concrete.Undo_Redoable;

with Debug; use Debug;
with Ada.Unchecked_Deallocation;

package body Logs is

   procedure Free_Stream is new Ada.Unchecked_Deallocation
     (Buffered_Stream_Type'Class, Buffered_Stream_Ref);

   procedure Free_Stream is new Ada.Unchecked_Deallocation
     (Stream_Type'Class, Stream_Ref);

   procedure Free_Buffer is new Ada.Unchecked_Deallocation
     (Unbounded_Buffer_Type'Class, Unbounded_Buffer_Ref);

   procedure Initialize (Log : in out Log_Type;
                         Recovery_Needed : out Boolean;
                         Anchor : in Stable_Params_Type'Class;
                         Params : in Stable_Params_Type'Class) is
      New_System : Boolean := False;
   begin
      Put_Line ("Log Initialization Started");
      Log.Params := new Stable_Params_Type'Class' (Params);
      begin
         Log.Anchor_Storage := Stable_Storage_Ref (Open_Storage (Anchor));
      exception
         when Storage_Not_Existant =>
            New_System := True;
      end;
      if New_System then
         Put_Line ("Log: Starting New System");
         Log.Anchor_Storage := Stable_Storage_Ref (Create_Storage (Anchor));
         Put_Line ("Log: Created Anchor");
         declare
            Temp_Buffer : Unbounded_Buffer_Ref :=
             new Unbounded_Buffer_Type (Log.Anchor_Storage);
            Temp_Stream : Buffered_Stream_Ref :=
             new Buffered_Stream_Type (Temp_Buffer);
         begin
            Log.Anchor.Status := Normal;
            Anchor_Type'Write (Temp_Stream, Log.Anchor);
            Flush (Temp_Stream.all);
            Close (Log.Anchor_Storage.all);
            Put_Line ("Log: Written Anchor");
            Free_Stream (Temp_Stream);
            Free_Buffer (Temp_Buffer);

            --  Prepare log for action
            Log.Log_Storage := Stable_Storage_Ref (Create_Storage (Params));
            Log.Log_Buffer := new Unbounded_Buffer_Type (Log.Log_Storage);
            Log.Log_Stream := new Buffered_Stream_Type (Log.Log_Buffer);
         exception
            when Storage_Existant =>
               raise Log_Corrupt;
         end;
         Recovery_Needed := False;
      else
         Put_Line ("Log: Old System");
         declare
            Temp_Buffer : Unbounded_Buffer_Ref :=
             new Unbounded_Buffer_Type (Log.Anchor_Storage);
            Temp_Stream : Buffered_Stream_Ref :=
             new Buffered_Stream_Type (Temp_Buffer);
         begin
            Anchor_Type'Read (Temp_Stream, Log.Anchor);
            Close (Log.Anchor_Storage.all);
            Free_Stream (Temp_Stream);
            Free_Buffer (Temp_Buffer);
         end;

         Put_Line ("Log: Read Anchor");
         Put_Line ("Log: Log status was " & System_Status_Type'Image
                   (Log.Anchor.Status));
         if Log.Anchor.Status = Normal then
            Recovery_Needed := False;

            --  Prepare log for action
            Log.Log_Storage := Stable_Storage_Ref
              (Open_Storage (Params, Out_Mode));
            Log.Log_Buffer := new Unbounded_Buffer_Type (Log.Log_Storage);
            Log.Log_Stream := new Buffered_Stream_Type (Log.Log_Buffer);

            declare
               Temp_Stream : Stream_Ref :=
                new Stream_Type (Log.Anchor_Storage);
            begin
               Open (Log.Anchor_Storage.all, Out_Mode);
               Log.Anchor.Status := Logs.Recovery_Needed;
               Anchor_Type'Write (Temp_Stream, Log.Anchor);
               Close (Log.Anchor_Storage.all);
               Put_Line ("Log: Set Anchor to recovery needed");
               Free_Stream (Temp_Stream);
            end;

         else
            Recovery_Needed := True;
         end if;
      end if;
   end Initialize;

   protected body Log_Protection_Type is

      procedure Log_Transaction_Status (Log : in out Log_Type;
                                        Transaction : in Transaction_Ref;
                                        Status : in Transaction_Status_Type) is
         Transaction_Information : Log_Information_Type'Class
           := Transaction_Information_Type'
           (Get_Simple_Transaction_Identifier (Transaction.all),
            Status);
      begin
         Put_Line ("Log: Setting Transaction"
                   & Transaction_Text (Transaction.all) & " to "
                   & Transaction_Status_Type'Image (Status));
         Log_Information_Type'Class'Output (Log.Log_Stream,
                                            Transaction_Information);
         Flush (Log.Log_Stream.all);
      end Log_Transaction_Status;

      procedure Log_Undo_Information (Log : in out Log_Type;
                                      Transaction : in Transaction_Ref;
                                      Object : in Memory_Object_Ref) is
      begin
         declare
            Undo_List : Logging_Technique_Array_Type :=
              Get_Undo_List (Undo_Redoable_Object_Ref (Object).all,
                             Transaction);
            Log_Info : Undo_Information_Type :=
              Create (Object, Undo_List);
         begin
            Put_Line ("Log: Write Undo for TA: "
                      & Transaction_Text (Transaction.all));
            Log_Info.Transaction :=
              Get_Simple_Transaction_Identifier (Transaction.all);
            Log_Information_Type'Class'Output
              (Log.Log_Stream, Log_Information_Type'Class (Log_Info));
            Flush (Log.Log_Stream.all);
         end;
      exception
         when Transaction_Not_Found =>
            Put_Line ("Log: This object has not been modified");
      end Log_Undo_Information;

      procedure Log_Redo_Information (Log : in out Log_Type;
                                      Transaction : in Transaction_Ref;
                                      Object : in Memory_Object_Ref) is
      begin
         declare
            Redo_List : Logging_Technique_Array_Type :=
              Get_Redo_List (Undo_Redoable_Object_Ref (Object).all,
                             Transaction);
            Log_Info : Redo_Information_Type :=
              Create (Object, Redo_List);
         begin
            Put_Line ("Log: Write Redo for TA: "
                      & Transaction_Text (Transaction.all));
            Log_Info.Transaction :=
              Get_Simple_Transaction_Identifier (Transaction.all);
            Log_Information_Type'Class'Output
              (Log.Log_Stream, Log_Information_Type'Class (Log_Info));
            Flush (Log.Log_Stream.all);
         end;
      exception
         when Transaction_Not_Found =>
            Put_Line ("Log: This object has not been modified");
      end Log_Redo_Information;

      procedure Log_Creation_Information
        (Log : in out Log_Type;
      Transaction : in Transaction_Ref;
      Params : in Non_Volatile_Params_Type'Class) is
         Log_Info : Create_Information_Type :=
           Create (Params);
      begin
         Put_Line ("Log: Write Creation for TA: "
                   & Transaction_Text (Transaction.all));
         Log_Info.Transaction :=
           Get_Simple_Transaction_Identifier (Transaction.all);
         Log_Information_Type'Class'Output
           (Log.Log_Stream, Log_Information_Type'Class (Log_Info));
         Flush (Log.Log_Stream.all);
      end Log_Creation_Information;

   end Log_Protection_Type;

   procedure Log_Transaction_Status (Log : in out Log_Type;
                                     Transaction : in Transaction_Ref;
                                     Status : in Transaction_Status_Type) is
   begin
      Log.Protection.Log_Transaction_Status (Log, Transaction, Status);
   end Log_Transaction_Status;

   procedure Log_Undo_Information (Log : in out Log_Type;
                                   Transaction : in Transaction_Ref;
                                   Object : in Memory_Object_Ref) is
   begin
      Log.Protection.Log_Undo_Information (Log, Transaction, Object);
   end Log_Undo_Information;

   procedure Log_Redo_Information (Log : in out Log_Type;
                                   Transaction : in Transaction_Ref;
                                   Object : in Memory_Object_Ref) is
   begin
      Log.Protection.Log_Redo_Information (Log, Transaction, Object);
   end Log_Redo_Information;

   procedure Log_Creation_Information
     (Log : in out Log_Type;
      Transaction : in Transaction_Ref;
      Params : in Non_Volatile_Params_Type'Class) is
   begin
      Log.Protection.Log_Creation_Information (Log, Transaction, Params);
   end Log_Creation_Information;

   procedure Start_Recovery (Log : in out Log_Type) is
   begin
      --  Prepare log for recovery
      Log.Log_Storage := Stable_Storage_Ref
        (Open_Storage (Log.Params.all, In_Mode));
      Log.Log_Buffer := new Unbounded_Buffer_Type (Log.Log_Storage);
      Log.Log_Stream := new Buffered_Stream_Type (Log.Log_Buffer);
   end Start_Recovery;

   procedure End_Recovery (Log : in out Log_Type) is
   begin
      --  Close log
      Close (Log.Log_Storage.all);
      --  Prepare log for action
      Open (Log.Log_Storage.all, Out_Mode);
   end End_Recovery;

   function Get_Next_Log_Information (Log : in Log_Type)
      return Log_Information_Type'Class is
   begin
      return Log_Information_Type'Class'Input (Log.Log_Stream);
   exception
      when others =>
         --  maybe dangerous
         raise No_More_Info;
   end Get_Next_Log_Information;

   procedure Shutdown (Log : in out Log_Type) is
   begin
      Put_Line ("Log: Shutting down system");
      declare
         Temp_Stream : Stream_Ref :=
          new Stream_Type (Log.Anchor_Storage);
      begin
         Open (Log.Anchor_Storage.all, Out_Mode);
         Log.Anchor.Status := Logs.Normal;
         Anchor_Type'Write (Temp_Stream, Log.Anchor);
         Close (Log.Anchor_Storage.all);
         Put_Line ("Log: Set Anchor to normal");
         Free_Stream (Temp_Stream);
      end;
   end Shutdown;

end Logs;


