-------------------------------------------------------------------------------
--                            Redo Recovery Manager                          --
--                                                                           --
--  This is the type for the redo recovery manager                           --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/25 13:58:32 $
--  $RCSfile: recovery_managers-redo.adb,v $
--  $Revision: 1.14 $
--                                                                           --
-------------------------------------------------------------------------------

with Debug; use Debug;

with Memory_Objects; use Memory_Objects;
with Lock_Managers; use Lock_Managers;

--  with Logging_Techniques; use Logging_Techniques;

with Log_Information; use Log_Information;
with Log_Information.Object; use Log_Information.Object;
with Log_Information.Transaction; use Log_Information.Transaction;
with Log_Information.Object.Redo; use Log_Information.Object.Redo;
with Log_Information.Object.Create; use Log_Information.Object.Create;

with Storage_Params.Non_Volatile; use Storage_Params.Non_Volatile;

with Cache_Managers; use Cache_Managers;
with Recovery_Support; use Recovery_Support;

with Keyed_Lists;
with Refkeyed_Lists;

package body Recovery_Managers.Redo is

   function Create (Log_Anchor  : in Stable_Params_Type'Class;
                    Log_Storage : in Stable_Params_Type'Class)
     return Redo_Recovery_Manager_Ref is
      Result : Redo_Recovery_Manager_Ref := new Redo_Recovery_Manager_Type;
   begin
      Initialize (Result.all, Log_Anchor, Log_Storage);
      return Result;
   end Create;

   procedure Pre_Operation (Manager : in out Redo_Recovery_Manager_Type;
                            Object  : in Concrete_Memory_Object_Ref;
                            Transaction : in Transaction_Ref) is
   begin
      Pin (Object.all);
   end Pre_Operation;

   procedure Post_Operation (Manager : in out Redo_Recovery_Manager_Type;
                             Object  : in Concrete_Memory_Object_Ref;
                             Transaction : in Transaction_Ref) is
   begin
      null;
   end Post_Operation;

   package Recovery_Lists is new Keyed_Lists
     (Simple_Transaction_Identifier_Type,
      Log_Information_Ref, Transaction_Status_Type);

   use Recovery_Lists;

   package Object_Recovery_Lists is new Refkeyed_Lists
     (Non_Volatile_Params_Type,
      Non_Volatile_Params_Ref, Log_Information_Ref, Natural);

   Object_Recovery_List : Object_Recovery_Lists.List_Type;

   procedure Handle_Commit (Object_List : Recovery_Lists.Element_Node_Ref);

   procedure Handle_Commit (Object_List : Recovery_Lists.Element_Node_Ref) is
      Tmp : Element_Node_Ref := Object_List;
   begin
      while Tmp /= null loop
         --  insert the object in the object recovery
         --  list, if necessary
         Put_Line ("Recovery: Added object", Initialization);
         Object_Recovery_Lists.Insert_Element_With_Info (Object_Recovery_List,
            Object_Information_Ref (Tmp.Element).Params,
            Tmp.Element, Object_Information_Ref (Tmp.Element)
            . State_Identifier, True);
         Tmp := Tmp.Next;
      end loop;
   end Handle_Commit;

   procedure Recover (Manager : in out Redo_Recovery_Manager_Type) is

      List : List_Type;

      Cache_Manager : Cache_Manager_Ref := Get_Cache_Manager;

      procedure Tag_Object (Params : Non_Volatile_Params_Ref);

      procedure Tag_Object (Params : Non_Volatile_Params_Ref) is
      begin
         Tag_Object (Cache_Manager.all, Params.all);
      end Tag_Object;

      procedure Tag_Objects is new
        Object_Recovery_Lists.Iterate_Through_Keys (Tag_Object);

      procedure Handle_Possible_Create
        (Transaction : in out Simple_Transaction_Identifier_Type);

      procedure Handle_Possible_Create
        (Transaction : in out Simple_Transaction_Identifier_Type) is
         Tmp : Element_Node_Ref := Find (List, Transaction);
      begin
         while Tmp /= null loop
            if Tmp.Element.all in Create_Information_Type'Class then
               Put_Line ("Recovery: Object created during transaction",
                         Initialization);
               Delete (Create_Information_Type'Class (Tmp.Element.all));
            end if;
            Tmp := Tmp.Next;
         end loop;
         Delete (List, Transaction);
      end Handle_Possible_Create;

      procedure Check_For_Creation is new
        Iterate_Through_Keys (Handle_Possible_Create);

   begin
      Put_Line ("Recovery: Starting recovery", Initialization);
      Start_Recovery (Manager.Log);
      begin
         loop
            declare
               Log_Info : Log_Information_Type'Class :=
                 Get_Next_Log_Information (Manager.Log);
            begin
               if Log_Info in Transaction_Information_Type'Class then
                  Put_Line ("Recovery: Transaction Info:" &
                            Transaction_Text (Log_Info.Transaction)
                            & " / State: "
                            & Transaction_Status_Type'Image
                            (Transaction_Information_Type (Log_Info) .Status),
                            Initialization);
                  case Transaction_Information_Type (Log_Info) .Status is
                     when Open =>
                        --  Insert new transaction into recovery list
                        Create_Key (List, Log_Info.Transaction, Open);
                     when Closed =>
                        null;
                     when Committed =>
                        Set_Info (List, Log_Info.Transaction, Committed);
                        declare
                           Object_List : Element_Node_Ref :=
                              Find (List, Log_Info.Transaction);
                        begin
                           if Object_List = null then
                              Put_Line ("Recovery: Transaction" &
                                 Transaction_Text (Log_Info.Transaction) &
                                 " needs no recovery", Initialization);
                              Delete (List, Log_Info.Transaction);
                           else
                              Put_Line ("Recovery: Transaction" &
                                 Transaction_Text (Log_Info.Transaction) &
                                 " needs recovery", Initialization);
                              Handle_Commit (Object_List);
                              Delete (List, Log_Info.Transaction);
                           end if;
                        end;
                     when Aborted =>
                        --  we handle aborted transaction at the end
                        null;
                  end case;
               elsif Log_Info in Redo_Information_Type'Class then
                  Put_Line ("Recovery: Redo Info for Transaction:" &
                            Transaction_Text (Log_Info.Transaction),
                            Initialization);
                  --  Insert Redo-information into recovery list
                  Insert_Element (List, Log_Info.Transaction,
                     new Log_Information_Type'Class' (Log_Info));
               elsif Log_Info in Create_Information_Type'Class then
                  Put_Line ("Recovery: Create Info for Transaction:" &
                            Transaction_Text (Log_Info.Transaction),
                            Initialization);
                  --  Insert Create-Info in recovery list
                  Insert_Element (List, Log_Info.Transaction,
                     new Log_Information_Type'Class' (Log_Info));
               end if;
            end;
         end loop;
      exception
         when No_More_Info =>
            null;
      end;
      End_Recovery (Manager.Log);

      --  Check if there have been newly created objects
      Check_For_Creation (List);

      Put_Line ("Recovery: Tagging objects", Initialization);
      Tag_Objects (Object_Recovery_List);
      Put_Line ("Recovery: Finished Recovery", Initialization);
   end Recover;

   procedure Recover_Object (Manager : in out Redo_Recovery_Manager_Type;
                             Object : in Concrete_Memory_Object_Ref;
                             Params : in Non_Volatile_Params_Ref;
                             Loading : in Loading_Operation_Type'Class) is

      use Object_Recovery_Lists;

      Current_State : Natural;
      Object_Redo_List : Object_Recovery_Lists.Element_Node_Ref;
      Tmp : Object_Recovery_Lists.Element_Node_Ref;
      N : Natural := 0;
   begin
      Current_State := Get_State_Identifier (Object.all);
      Put_Line ("Recovery: Recover_Object called: current state"
                & Natural'Image (Current_State), Initialization);
      Object_Redo_List := Object_Recovery_Lists.Find
        (Object_Recovery_List, Params);
      Tmp := Object_Redo_List;
      --  count the number of redo operations that must be done
      while Tmp /= null loop
         Put_Line ("Recovery: Redo_Information state:" &
                   Natural'Image (Object_Information_Ref (Tmp.Element)
                                  .State_Identifier), Initialization);
         if Object_Information_Ref (Tmp.Element) .State_Identifier >=
           Current_State then
            N := N + 1;
         end if;
         Tmp := Tmp.Next;
      end loop;
      if N > 0 then
         declare
            Operation_List : array (1 .. N) of Redo_Information_Ref;
            I : Integer := 1;
         begin
            Tmp := Object_Redo_List;
            while I <= N loop
               Operation_List (I) := Redo_Information_Ref (Tmp.Element);
               I := I + 1;
               Tmp := Tmp.Next;
            end loop;
            --  now redo all operations
            --  we must start at the end of the list
            for I in reverse Operation_List'Range loop
               Put_Line ("Recovery: Redoing operation",
                         Initialization);
               Log_Information.Object.Redo.Redo
                 (Info => Operation_List (I).all,
                  Object => Memory_Object_Ref (Object),
                  Loading => Loading);
            end loop;
         end;
      else
         Put_Line ("Recovery: Nothing to recover for this object",
                   Initialization);
      end if;
      Object_Recovery_Lists.Delete (Object_Recovery_List, Params);
   end Recover_Object;

   procedure Abort_Transaction (Manager : in out Redo_Recovery_Manager_Type;
                                Transaction : in Transaction_Ref) is
      Objects : Accessed_Objects_Type :=
        Get_Accessed_Objects (Transaction.all);
      Is_Toplevel : Boolean := Transactions.Is_Toplevel (Transaction.all);
   begin
      for I in Objects'Range loop
         Abort_Transaction (Concrete_Memory_Object_Ref (Objects (I)) .all,
                            Transaction);
         if Is_Toplevel then
            Un_Pin (Concrete_Memory_Object_Ref (Objects (I)) .all);
         end if;
      end loop;

      Log_Transaction_Status (Manager.Log, Transaction, Aborted);

      for I in Objects'Range loop
         --  handle the locks
         Transaction_Abort (Get_Lock_Manager (Concrete_Memory_Object_Ref
            (Objects (I)).all).all, Transaction);
      end loop;

   end Abort_Transaction;

   procedure Commit_Transaction (Manager : in out Redo_Recovery_Manager_Type;
                                 Transaction : in Transaction_Ref) is
      Objects : Accessed_Objects_Type :=
        Get_Modified_Objects (Transaction.all);
   begin
      Log_Transaction_Status (Manager.Log, Transaction, Closed);

      Put_Line ("Recovery: Transaction accessed" & Integer'Image
                (Objects'Length));

      if Is_Toplevel (Transaction.all) then
         for I in Objects'Range loop
            --  write the Redo information for each object to disk
            Log_Redo_Information (Manager.Log, Transaction, Objects (I));
         end loop;
      end if;

      Log_Transaction_Status (Manager.Log, Transaction, Committed);

      --  go through all accessed objects
      for I in Objects'Range loop
         --  handle the locks
         Commit_Transaction (Concrete_Memory_Object_Ref (Objects (I)).all,
                             Transaction);
         Transaction_Commit (Get_Lock_Manager
            (Concrete_Memory_Object_Ref (Objects (I)).all).all, Transaction);
      end loop;

      if Is_Toplevel (Transaction.all) then
         for I in Objects'Range loop
            --  finally unpin the object
            Un_Pin (Concrete_Memory_Object_Ref (Objects (I)) .all);
            --  propagate to the disk, if necessary
            Propagate (Concrete_Memory_Object_Ref (Objects (I)) .all);
         end loop;
      end if;

   end Commit_Transaction;

end Recovery_Managers.Redo;



