-------------------------------------------------------------------------------
--                            Undo Recovery Manager                          --
--                                                                           --
--  This is the type for the Undo recovery manager                           --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/23 11:19:12 $
--  $RCSfile: recovery_managers-undo.adb,v $
--  $Revision: 1.4 $
--                                                                           --
-------------------------------------------------------------------------------

with Debug; use Debug;

with Memory_Objects; use Memory_Objects;

with Lock_Managers; use Lock_Managers;

with Log_Informations; use Log_Informations;
with Log_Informations.Transaction; use Log_Informations.Transaction;
with Log_Informations.Object.Undo; use Log_Informations.Object.Undo;
with Log_Informations.Object.Redo; use Log_Informations.Object.Redo;

with Ada.Finalization;

package body Recovery_Managers.Undo is

   type Node_Type;

   type Node_Ref is access Node_Type;

   type Node_Type is record
      Next : Node_Ref;
      Prev : Node_Ref;
      Transaction : Transaction_Ref;
      Element : Element_Type;
   end record;

   type Recovery_List_Type is new Limited_Controlled with record
      First : Node_Ref;
      Last : Node_Ref;
   end record;

   function Find (Recovery_List : Recovery_List_Type;
                  Transaction : Transaction_Ref) return Element_Type;

   procedure Insert (Recovery_List : in out Recovery_List_Type;
                     Transaction : in Transaction_Ref;
                     Element : in Element_Type);

   procedure Append (Recovery_List : in out Recovery_List_Type;
                     Transaction : in Transaction_Ref;
                     Element : in Logging_Technique_List_Ref);

   procedure Delete (Recovery_List : in out Recovery_List_Type;
                     Transaction : in Transaction_Ref);

   Transaction_Not_Found,
     Transaction_Already_Inserted : exception;

   procedure Finalize (Recovery_List : in out Recovery_List_Type);

   procedure Free_Node is new Ada.Unchecked_Deallocation
     (Transaction_Node_Type, Transaction_Node_Ref);

   function Find (Recovery_List : Recovery_List_Type;
                  Transaction : Transaction_Ref) return Undo_List_Type is
      Tmp : Transaction_Node_Ref := Recovery_List.First;
   begin
      while Tmp /= null loop
         if Tmp.Transaction = Transaction then
            return Tmp.List;
         else
            Tmp := Tmp.Next;
         end if;
      end loop;
      raise Transaction_Not_Found;
   end Find;

   procedure Insert (Recovery_List : in out Recovery_List_Type;
                     Transaction : in Transaction_Ref;
                     List : in Undo_List_Type) is
      Tmp : Transaction_Node_Ref := Recovery_List.First;
   begin
      while Tmp /= null loop
         if Tmp.Transaction = Transaction then
            raise Transaction_Already_Inserted;
         else
            Tmp := Tmp.Next;
         end if;
      end loop;
      --  We insert it at the beginning
      declare
         New_Node : Transaction_Node_Ref := new Transaction_Node_Type;
      begin
         New_Node.Next := Recovery_List.First;
         New_Node.Transaction := Transaction;
         New_Node.List := List;
         if Recovery_List.First /= null then
            Recovery_List.First.Prev := New_Node;
         else
            Recovery_List.Last := New_Node;
         end if;
         Recovery_List.First := New_Node;
      end;
   end Insert;

   procedure Append (Recovery_List : in out Recovery_List_Type;
                     Transaction : in Transaction_Ref;
                     Element : in Logging_Technique_List_Ref) is
      Tmp : Transaction_Node_Ref := Recovery_List.First;
   begin
      while Tmp /= null loop
         if Tmp.Transaction = Transaction then
            Element.Next := Tmp.List.First;
            Tmp.List.First := Element;
            return;
         else
            Tmp := Tmp.Next;
         end if;
      end loop;
      raise Transaction_Not_Found;
   end Append;

   procedure Delete (Recovery_List : in out Recovery_List_Type;
                     Transaction : in Transaction_Ref) is
      Tmp : Transaction_Node_Ref := Recovery_List.First;
   begin
      while Tmp /= null loop
         if Tmp.Transaction = Transaction then
            if Tmp.Prev = null then
               Recovery_List.First := Tmp.Next;
            else
               Tmp.Prev.Next := Tmp.Next;
            end if;
            if Tmp.Next = null then
               Recovery_List.Last := Tmp.Prev;
            else
               Tmp.Next.Prev := Tmp.Prev;
            end if;
            Free_Node (Tmp);
            return;
         else
            Tmp := Tmp.Next;
         end if;
      end loop;
      raise Transaction_Not_Found;
   end Delete;

   procedure Finalize (Recovery_List : in out Recovery_List_Type) is
      Tmp : Transaction_Node_Ref := Recovery_List.First;
      To_Free : Transaction_Node_Ref;
   begin
      while Tmp /= null loop
         To_Free := Tmp;
         Tmp := Tmp.Next;
         Free_Node (To_Free);
      end loop;
   end Finalize;



   function Create (Log_Anchor  : in Stable_Params_Type'Class;
                    Log_Storage : in Stable_Params_Type'Class)
     return Undo_Recovery_Manager_Ref is
      Result : Undo_Recovery_Manager_Ref := new Undo_Recovery_Manager_Type;
   begin
      Initialize (Result.all, Log_Anchor, Log_Storage);
      return Result;
   end Create;

   procedure Pre_Operation (Manager : in out Undo_Recovery_Manager_Type;
                            Object  : in Concrete_Memory_Object_Ref;
                            Transaction : in Transaction_Ref) is
   begin
      Pin (Object.all);
   end Pre_Operation;

   procedure Post_Operation (Manager : in out Undo_Recovery_Manager_Type;
                             Object  : in Concrete_Memory_Object_Ref;
                             Transaction : in Transaction_Ref) is
   begin
      Un_Pin (Object.all);
   end Post_Operation;

   procedure Recover (Manager : in out Undo_Recovery_Manager_Type) is
   begin
      Put_Line ("Recovery: Starting recovery");
      Start_Recovery (Manager.Log);
      begin
         loop
            declare
               Log_Info : Log_Information_Type'Class :=
                 Get_Next_Log_Information (Manager.Log);
            begin
               if Log_Info in Transaction_Information_Type'Class then
                  Put_Line ("Recovery: Transaction Info:" &
                            Transaction_Text (Log_Info.Transaction)
                            & " / State: "
                            & Transaction_Status_Type'Image
                            (Transaction_Information_Type (Log_Info) .Status));
                  case Transaction_Information_Type (Log_Info) .Status is
                     when Active =>
                        --  Insert new transaction into recovery list
                        null;
                     when Preparing =>
                        null;
                     when Committed =>
                        null;
                     when Aborted =>
                        null;
                  end case;
               elsif Log_Info in Undo_Information_Type'Class then
                  Put_Line ("Recovery: Undo Info for Transaction:" &
                            Transaction_Text (Log_Info.Transaction));
                  --  Insert undo-information into recovery list
                  null;
               elsif Log_Info in Redo_Information_Type'Class then
                  Put_Line ("Recovery: Redo Info for Transaction:" &
                            Transaction_Text (Log_Info.Transaction));
                  --  Insert undo-information into recovery list
                  null;
               end if;
            end;
         end loop;
      exception
         when No_More_Info =>
            null;
      end;
      End_Recovery (Manager.Log);
      Put_Line ("Recovery: Finished Recovery");
   end Recover;

   procedure Abort_Transaction (Manager : in out Undo_Recovery_Manager_Type;
                                Transaction : in Transaction_Ref) is
      Objects : Accessed_Objects_Type :=
        Get_Accessed_Objects (Transaction.all);
   begin
      for I in Objects'Range loop
         Abort_Transaction (Concrete_Memory_Object_Ref (Objects (I)) .all,
                            Transaction);
      end loop;

      Log_Transaction_Status (Manager.Log, Transaction, Aborted);

      for I in Objects'Range loop
         --  handle the locks
         Transaction_Abort (Get_Lock_Manager (Concrete_Memory_Object_Ref
            (Objects (I)).all).all, Transaction);
      end loop;

   end Abort_Transaction;

   procedure Commit_Transaction (Manager : in out Undo_Recovery_Manager_Type;
                                 Transaction : in Transaction_Ref) is
      Objects : Accessed_Objects_Type :=
        Get_Accessed_Objects (Transaction.all);
   begin
      Log_Transaction_Status (Manager.Log, Transaction, Preparing);

      if Is_Toplevel (Transaction.all) then
         for I in Objects'Range loop
            --  write the undo information for each object to disk
            Log_Undo_Information (Manager.Log, Transaction, Objects (I));
            --  propagate to the disk, if necessary
            Propagate (Concrete_Memory_Object_Ref (Objects (I)) .all);
         end loop;
      end if;

      Set_Final_Transaction_Status (Transaction.all, Committed);

      --  go through all accessed objects
      for I in Objects'Range loop
         --  handle the locks
         Commit_Transaction (Concrete_Memory_Object_Ref (Objects (I)).all,
                             Transaction);
         Transaction_Commit (Get_Lock_Manager
            (Concrete_Memory_Object_Ref (Objects (I)).all).all, Transaction);
      end loop;

      Log_Transaction_Status (Manager.Log, Transaction, Committed);
   end Commit_Transaction;

end Recovery_Managers.Undo;
