-------------------------------------------------------------------------------
--                            Mirrored Storage Params                        --
--                                                                           --
--  Author: Xavier Caron                                                     --
--                                                                           --
--  This package provides the storage parameters for the mirrored            --
--  storage type.                                                            --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/08/18 11:48:50 $
--  $Revision: 1.3 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Storage_Params.Non_Volatile.Non_Stable.File_Storage_Params;
use  Storage_Params.Non_Volatile.Non_Stable.File_Storage_Params;
with Storages.Non_Volatile.Stable.Mirrored_Storages;
use  Storages.Non_Volatile.Stable.Mirrored_Storages;

package body Storage_Params.Non_Volatile.Stable.Mirrored_Storage_Params is

   ---------------------------------------------------------------------------

   function String_To_Storage_Params (S : in String)
     return Mirrored_Storage_Params_Type
   is
      P : Mirrored_Storage_Params_Type;
   begin
      P.Mirrored_Storage_Name := (N => S'Length, Name => S);
      P.Members.Data_1_Params := new File_Storage_Params_Type'
         (String_To_Storage_Params (S & "1.data"));

      P.Members.Data_2_Params := new File_Storage_Params_Type'
        (String_To_Storage_Params (S & "2.data"));

      P.Members.Log_Params := new File_Storage_Params_Type'
        (String_To_Storage_Params (S & ".log"));
      return P;
   end String_To_Storage_Params;

   function  String_To_Storage_Params
     (S : in String; Data_1_Params, Data_2_Params,
                       Log_Params : in Non_Volatile_Params_Type'Class)
      return Mirrored_Storage_Params_Type
   is
      P : Mirrored_Storage_Params_Type;
   begin
      P.Mirrored_Storage_Name := (N => S'Length, Name => S);
      P.Members.Data_1_Params := new Non_Volatile_Params_Type'Class'
        (Data_1_Params);
      P.Members.Data_2_Params := new Non_Volatile_Params_Type'Class'
        (Data_2_Params);
      P.Members.Log_Params := new Non_Volatile_Params_Type'Class'
        (Log_Params);
      return P;
   end String_To_Storage_Params;

   ---------------------------------------------------------------------------

   function Storage_Params_To_String (Params : in Mirrored_Storage_Params_Type)
                                      return String
   is
   begin
      return Params.Mirrored_Storage_Name.Name;
   end Storage_Params_To_String;

   ---------------------------------------------------------------------------

   function Get_Data_1_Params (Params : in Mirrored_Storage_Params_Type)
                               return Non_Volatile_Params_Type'Class
   is
   begin
      return Params.Members.Data_1_Params.all;
   end Get_Data_1_Params;

   function Get_Data_2_Params (Params : in Mirrored_Storage_Params_Type)
                               return Non_Volatile_Params_Type'Class
   is
   begin
      return Params.Members.Data_2_Params.all;
   end Get_Data_2_Params;

   function Get_Log_Params (Params : in Mirrored_Storage_Params_Type)
                            return Non_Volatile_Params_Type'Class
   is
   begin
      return Params.Members.Log_Params.all;
   end Get_Log_Params;

   ---------------------------------------------------------------------------

   function Create_Storage (Params : in Mirrored_Storage_Params_Type)
                            return Storage_Ref
   is
   begin
      return Storage_Ref (Create (Params));
   end Create_Storage;

   ---------------------------------------------------------------------------

   function Open_Storage (Params : in Mirrored_Storage_Params_Type;
                          Mode   : in Mode_Type := In_Mode)
                          return Non_Volatile_Storage_Ref
   is
   begin
      return Non_Volatile_Storage_Ref (Open (Params, Mode));
   end Open_Storage;

   ---------------------------------------------------------------------------

   procedure Free is
      new Ada.Unchecked_Deallocation (Non_Volatile_Params_Type'Class,
                                      Non_Volatile_Params_Ref);

   procedure Finalize (P : in out Mirrored_Params_Members_Type)
   is
   begin
      Free (P.Data_1_Params);
      Free (P.Data_2_Params);
      Free (P.Log_Params);
   end Finalize;

   procedure Adjust (P : in out Mirrored_Params_Members_Type)
   is
   begin
      P.Data_1_Params := new Non_Volatile_Params_Type'Class'
        (P.Data_1_Params.all);
      P.Data_2_Params := new Non_Volatile_Params_Type'Class'
        (P.Data_2_Params.all);
      P.Log_Params := new Non_Volatile_Params_Type'Class'
        (P.Log_Params.all);
   end Adjust;

   ---------------------------------------------------------------------------

   procedure My_Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                       Item   : in Mirrored_Params_Members_Type) is
   begin
      Non_Volatile_Params_Type'Class'Output (Stream, Item.Data_1_Params.all);
      Non_Volatile_Params_Type'Class'Output (Stream, Item.Data_2_Params.all);
      Non_Volatile_Params_Type'Class'Output (Stream, Item.Log_Params.all);
   end My_Write;

   procedure My_Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                      Item   : out Mirrored_Params_Members_Type) is
   begin
      Item.Data_1_Params := new Non_Volatile_Params_Type'Class'
        (Non_Volatile_Params_Type'Class'Input (Stream));
      Item.Data_2_Params := new Non_Volatile_Params_Type'Class'
        (Non_Volatile_Params_Type'Class'Input (Stream));
      Item.Log_Params := new Non_Volatile_Params_Type'Class'
        (Non_Volatile_Params_Type'Class'Input (Stream));
   end My_Read;

   ---------------------------------------------------------------------------

end Storage_Params.Non_Volatile.Stable.Mirrored_Storage_Params;
