-------------------------------------------------------------------------------
--                               File Storages                               --
--                                                                           --
--  This package is an implementation of a non-volatile, non-stable storage  --
--  based on the local file system. It uses the standard Stream_IO inter-    --
--  face.                                                                    --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/08/18 11:48:51 $
--  $RCSfile: storages-non_volatile-non_stable-file_storages.adb,v $
--  $Revision: 1.3 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Streams.Stream_IO; use Ada.Streams.Stream_IO;

package body Storages.Non_Volatile.Non_Stable.File_Storages is

   procedure Finalize (Data : in out Stream_Data) is
   begin
      if Data.Status = Connected then
         Close (Data.File);
         Data.Status := Not_Connected;
      end if;
   end Finalize;

   function Create (Params : in File_Storage_Params_Type)
                    return File_Storage_Ref is
      S : File_Storage_Ref := new File_Storage_Type;
   begin

      declare
      begin
         Stream_IO.Open (S.Data.File, In_File,
                         Storage_Params_To_String (Params));
         Stream_IO.Close (S.Data.File);
         raise Storage_Existant;
      exception
         when Name_Error =>
            null;
      end;

      Stream_IO.Create (S.Data.File, Out_File,
                        Storage_Params_To_String (Params));
      S.Data.Params := Params;
      S.Data.Status := Connected;
      return S;
   end Create;

   function Open (Params : in File_Storage_Params_Type;
                  Mode   : in Mode_Type := In_Mode)
                  return File_Storage_Ref is
      S : File_Storage_Ref := new File_Storage_Type;
   begin
      case Mode is
         when In_Mode =>
            Stream_IO.Open (S.Data.File, In_File,
                            Storage_Params_To_String (Params));
         when Out_Mode =>
            Stream_IO.Open (S.Data.File, Out_File,
                            Storage_Params_To_String (Params));
         when In_Out_Mode =>
            raise Mode_Not_Supported;
      end case;
      S.Data.Params := Params;
      S.Data.Status := Connected;
      return S;
   exception
      when Name_Error =>
         raise Storage_Not_Existant;
   end Open;

   procedure Open (Storage : in out File_Storage_Type;
                   Mode    : in Mode_Type := In_Mode) is
   begin
      case Mode is
         when In_Mode =>
            Open (Storage.Data.File, In_File,
                  Storage_Params_To_String (Storage.Data.Params));
         when Out_Mode =>
            Open (Storage.Data.File, Out_File,
                  Storage_Params_To_String (Storage.Data.Params));
         when In_Out_Mode =>
            raise Mode_Not_Supported;
      end case;
      Storage.Data.Status := Connected;
   end Open;

   procedure Close (Storage : in out File_Storage_Type) is
   begin
      Close (Storage.Data.File);
      Storage.Data.Status := Not_Connected;
   end Close;

   procedure Delete (Storage : in out File_Storage_Type) is
   begin
      Delete (Storage.Data.File);
      Storage.Data.Status := Not_Connected;
   end Delete;

   procedure Read (Storage : in out File_Storage_Type;
                   Item    : out Stream_Element_Array;
                   Last    : out Stream_Element_Offset) is
   begin
      Read (Storage.Data.File, Item, Last);
   exception
      when End_Error =>
         raise Data_Not_Available;
   end Read;

   procedure Write (Storage : in out File_Storage_Type;
                    Item    : in Stream_Element_Array) is
   begin
      Write (Storage.Data.File, Item);
      Flush (Storage.Data.File);
   end Write;

   function Get_Current_Size (Storage : in File_Storage_Type)
                              return Stream_Element_Count is
   begin
      return Stream_Element_Count (Size (Storage.Data.File));
   end Get_Current_Size;

end Storages.Non_Volatile.Non_Stable.File_Storages;
