-------------------------------------------------------------------------------
--                               File Storages                               --
--                                                                           --
--  This package is an implementation of a non-volatile, non-stable storage  --
--  based on the local file system. It uses the standard Stream_IO inter-    --
--  face.                                                                    --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/01/05 12:57:07 $
--  $RCSfile: storages-non_volatile-non_stable-file_storages.ads,v $
--  $Revision: 1.1.1.1 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Streams; use Ada.Streams;
with Ada.Finalization; use Ada.Finalization;
with Ada.Streams.Stream_IO;
with Storage_Params; use Storage_Params;
with Storage_Params.Non_Volatile.Non_Stable.File_Storage_Params;
use  Storage_Params.Non_Volatile.Non_Stable.File_Storage_Params;

package Storages.Non_Volatile.Non_Stable.File_Storages is

   type File_Storage_Type is new Non_Stable_Storage_Type with private;
   type File_Storage_Ref is access all File_Storage_Type'Class;

   function Create (Params : File_Storage_Params_Type) return File_Storage_Ref;
   --  Create a file identified by Params and a File_Storage_Type object,
   --  associate them and hand back a reference.

   function Open (Params : in File_Storage_Params_Type;
                  Mode   : in Mode_Type := In_Mode)
                  return File_Storage_Ref;
   --  Open an already existing file identified by Params, create a new
   --  File_Storage_Type object and associate them.

   procedure Read (Storage : in out File_Storage_Type;
                   Item    : out Stream_Element_Array;
                   Last    : out Stream_Element_Offset);
   --  Read a data item from the storage

   procedure Write (Storage : in out File_Storage_Type;
                    Item    : in Stream_Element_Array);
   --  Append 'Data' to the storage

   procedure Open (Storage : in out File_Storage_Type;
                   Mode    : in Mode_Type := In_Mode);

   procedure Close (Storage : in out File_Storage_Type);

   procedure Delete (Storage : in out File_Storage_Type);

   function Get_Current_Size (Storage : in File_Storage_Type)
                              return Stream_Element_Count;

private

   type Status_Type is (Not_Connected, Connected);

   type Stream_Data is new Limited_Controlled with record
      File   : Ada.Streams.Stream_IO.File_Type;
      Params : File_Storage_Params_Type;
      Status : Status_Type := Not_Connected;
   end record;

   --  Encapsulating the data of the File storage in a controlled type
   --  gives us the possibility to close the stream's File when it ceases to
   --  exist. This is very useful to avoid dangling open files in case of
   --  corrupted streams, where 'Read will raise exceptions.

   procedure Finalize
     (Data : in out Stream_Data);

   type File_Storage_Type is new Non_Stable_Storage_Type with
      record
         Data : Stream_Data;
      end record;

end Storages.Non_Volatile.Non_Stable.File_Storages;



