-------------------------------------------------------------------------------
--                              Mirrored Storages                            --
--                                                                           --
--  Stable storage types represent storage that can survive program          --
--  termination as well as crash failures. Using a technique called          --
--  mirroring , all or nothing semantics must be guaranteed                  --
--  for the Write procedure.                                                 --
--  The Mirrored Storage Type does mirroring with any non-volatile storage.  --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/08/18 11:48:52 $
--  $RCSfile: storages-non_volatile-stable-mirrored_storages.adb,v $
--  $Revision: 1.2 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Storages.Non_Volatile; use Storages.Non_Volatile;
with Storage_Params.Non_Volatile; use Storage_Params.Non_Volatile;
with Storage_Params.Non_Volatile.Stable.Mirrored_Storage_Params;
use Storage_Params.Non_Volatile.Stable.Mirrored_Storage_Params;

package body Storages.Non_Volatile.Stable.Mirrored_Storages is

   ---------------------------------------------------------------------------

   procedure Free is new Ada.Unchecked_Deallocation
     (Non_Volatile_Storage_Type'Class, Non_Volatile_Storage_Ref);

   procedure Finalize (Members : in out Mirrored_Members_Type)
   is
   begin
      Free (Members.MirrorFile1);
      Free (Members.MirrorFile2);
      Free (Members.LogFile);
   end Finalize;

   ---------------------------------------------------------------------------
   --  The procedure Cleanup copies the content of one mirrored data location
   --  (chosen with the parameter Original) on the other one. If there is a
   --  crash during the cleanup, the original data location is garanted to
   --  remain intact.

   procedure Cleanup (Storage : in out Mirrored_Storage_Type;
                      Original : in Duplicata)
   is
      Size : Stream_Element_Count;
      Block : Stream_Element_Array (1 .. 1024);
      Last : Stream_Element_Offset;
   begin
      case Original is
         when 1 =>
            Storage.Members.MirrorFile1 :=
              Open_Storage (Get_Data_1_Params
                            (Storage.Members.Params), In_Mode);
            Storage.Members.MirrorFile2 :=
              Open_Storage (Get_Data_2_Params
                            (Storage.Members.Params), Out_Mode);
            Size := Get_Current_Size (Storage.Members.MirrorFile1.all);

            while Size >= 1024 loop
               Read (Storage.Members.MirrorFile1.all, Block, Last);
               Write (Storage.Members.MirrorFile2.all, Block);
               Size := Size - 1024;
            end loop;
            if Size > 0 then
               Read (Storage.Members.MirrorFile1.all, Block, Last);
               Write (Storage.Members.MirrorFile2.all, Block (1 .. Size));
            end if;

         when 2 =>
            Storage.Members.MirrorFile1 :=
              Open_Storage (Get_Data_1_Params
                            (Storage.Members.Params), Out_Mode);
            Storage.Members.MirrorFile2 :=
              Open_Storage (Get_Data_2_Params
                            (Storage.Members.Params), In_Mode);
            Size := Get_Current_Size (Storage.Members.MirrorFile2.all);

            while Size >= 1024 loop
               Read (Storage.Members.MirrorFile2.all, Block, Last);
               Write (Storage.Members.MirrorFile1.all, Block);
               Size := Size - 1024;
            end loop;
            if Size > 0 then
               Read (Storage.Members.MirrorFile2.all, Block, Last);
               Write (Storage.Members.MirrorFile1.all, Block (1 .. Size));
            end if;
      end case;
      Close (Storage.Members.MirrorFile1.all);
      Close (Storage.Members.MirrorFile2.all);

   end Cleanup;

   ---------------------------------------------------------------------------
   --  We declare the procedure Verify here just to compile with -gnatg.

   procedure Verify (Storage : in out Mirrored_Storage_Ref);

   --  The procedure Verify checks with the log if we recover of some
   --  crash failure or something unnormal and determines if there is some
   --  cleanup to launch.

   procedure Verify (Storage : in out Mirrored_Storage_Ref)
   is
      Logmessage : Stream_Element_Array (1 .. 2);
      --  Logmessage has 2 elements so that when we read the log, last
      --  will be equal to 2 if there is too much elements in the log location.
      --  (considered as corrupted).
      Last : Stream_Element_Offset := 0;
   begin
      declare
      begin
         Storage.Members.LogFile :=
           Open_Storage (Get_Log_Params
                         (Storage.Members.Params), In_Mode);
         --  Try to read log location.
         Read (Storage.Members.LogFile.all, Logmessage, Last);
         if Last /= 1 then
            raise Data_Not_Available;
         end if;

         case Logmessage (1) is
            when 0 =>
               return; --  Everything is OK
            when 1 =>
               --  Crash during first write (in data1)
               --  Cleanup based upon data2.
               Cleanup (Storage.all, 2);
            when 2 =>
               --  Crash during second write (in data2)
               --  Cleanup based upon data1.
               Cleanup (Storage.all, 1);
            when others =>  --  oops, should not happen!
               raise Data_Not_Available;
         end case;

         --  Cleanup seems to have succeeded,
         --  open the log again, but in Out_mode,
         --  and write OK into it.
         Close (Storage.Members.LogFile.all);
         Open (Storage.Members.LogFile.all, Out_Mode);
         Logmessage (1) := 0;
         Write (Storage.Members.LogFile.all, Logmessage (1 .. 1));

      exception
         when Data_Not_Available =>

            --  If log is unreadable, then we choose to copy
            --  the first data location on the second one.
            Cleanup (Storage.all, 1);

            --  Cleanup seems to have succeeded,
            --  open the log location again, but in Out_mode,
            --  and write OK into it.
            Close (Storage.Members.LogFile.all);
            Open (Storage.Members.LogFile.all, Out_Mode);
            Logmessage (1) := 0;
            Write (Storage.Members.LogFile.all, Logmessage (1 .. 1));
      end; --  declare

   end Verify;

   ---------------------------------------------------------------------------

   function Create (Params : in Mirrored_Storage_Params_Type)
                    return Mirrored_Storage_Ref
   is
      M : Mirrored_Storage_Ref := new Mirrored_Storage_Type;
      Logmessage : Stream_Element_Array (1 .. 1);
   begin
      M.Members.Params := Params;
      M.Members.MirrorFile1 := Non_Volatile_Storage_Ref
        (Create_Storage  (Get_Data_1_Params (Params)));
      M.Members.MirrorFile2 := Non_Volatile_Storage_Ref
        (Create_Storage (Get_Data_2_Params (Params)));
      M.Members.LogFile := Non_Volatile_Storage_Ref
        (Create_Storage (Get_Log_Params (Params)));
      Logmessage (1) := 0;
      Write (M.Members.LogFile.all, Logmessage);
      return M;
   end Create;

   ---------------------------------------------------------------------------

   function Open (Params : in Mirrored_Storage_Params_Type;
                  Mode : in Mode_Type := In_Mode)
      return Mirrored_Storage_Ref is
      M : Mirrored_Storage_Ref := new Mirrored_Storage_Type;
   begin
      M.Members.Params := Params;
      Verify (M); --  check if we recover from crash failure. If yes, cleanup.
      case Mode is
         when In_Out_Mode =>
            raise Mode_Not_Supported;
         when others =>
            M.Members.MirrorFile1 := Open_Storage
              (Get_Data_1_Params (Params), Mode);
            M.Members.MirrorFile2 := Open_Storage
              (Get_Data_2_Params (Params), Mode);
      end case;
      return M;
   end Open;

   ---------------------------------------------------------------------------

   procedure Open (Storage : in out Mirrored_Storage_Type;
                   Mode : in Mode_Type := In_Mode) is
   begin
      Open (Storage.Members.MirrorFile1.all, Mode);
      Open (Storage.Members.MirrorFile2.all, Mode);
      Open (Storage.Members.LogFile.all, Out_Mode);
   end Open;

   ---------------------------------------------------------------------------

   procedure Close (Storage : in out Mirrored_Storage_Type) is
   begin
      Close (Storage.Members.LogFile.all);
      Close (Storage.Members.MirrorFile1.all);
      Close (Storage.Members.MirrorFile2.all);
   end Close;

   ---------------------------------------------------------------------------

   procedure Delete (Storage : in out Mirrored_Storage_Type) is
   begin
      Delete (Storage.Members.MirrorFile1.all);
      Delete (Storage.Members.MirrorFile2.all);
      Delete (Storage.Members.LogFile.all);
   end Delete;

   ---------------------------------------------------------------------------

   procedure Read (Storage : in out Mirrored_Storage_Type;
                   Item : out Stream_Element_Array;
                   Last : out Stream_Element_Offset) is
   begin
      Read (Storage.Members.MirrorFile1.all, Item, Last);
   end Read;

   ---------------------------------------------------------------------------

   procedure Write (Storage : in out Mirrored_Storage_Type;
                    Item : in Stream_Element_Array) is
      Logmessage : Stream_Element_Array (1 .. 1);
   begin
      Logmessage (1) := 1;  --  We start the first write
      Close (Storage.Members.LogFile.all);
      Open (Storage.Members.LogFile.all, Out_Mode);
      Write (Storage.Members.LogFile.all, Logmessage);

      Write (Storage.Members.MirrorFile1.all, Item);

      Logmessage (1) := 2;  --  First write is successfully completed.
      --  We start the second one.
      Close (Storage.Members.LogFile.all);
      Open (Storage.Members.LogFile.all, Out_Mode);
      Write (Storage.Members.LogFile.all, Logmessage);

      Write (Storage.Members.MirrorFile2.all, Item);

      Logmessage (1) := 0;  --  Second write is successfully completed.
      --  Everything is OK.
      Close (Storage.Members.LogFile.all);
      Open (Storage.Members.LogFile.all, Out_Mode);
      Write (Storage.Members.LogFile.all, Logmessage);
   end Write;

   ---------------------------------------------------------------------------

   function Get_Current_Size (Storage : in Mirrored_Storage_Type)
                              return Stream_Element_Count is
   begin
      return Get_Current_Size (Storage.Members.MirrorFile1.all);
   end Get_Current_Size;

   ---------------------------------------------------------------------------

end Storages.Non_Volatile.Stable.Mirrored_Storages;
