-------------------------------------------------------------------------------
--                              Mirrored Storages                            --
--                                                                           --
--  Stable storage types represent storage that can survive program          --
--  termination as well as crash failures. Using a technique called          --
--  mirroring , all or nothing semantics must be guaranteed                  --
--  for the Write procedure.                                                 --
--  The Mirrored Storage Type does mirroring with any non-volatile storage.  --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/01/05 12:57:07 $
--  $RCSfile: storages-non_volatile-stable-mirrored_storages.ads,v $
--  $Revision: 1.1.1.1 $
--                                                                           --
-------------------------------------------------------------------------------

with Storage_Params.Non_Volatile.Stable.Mirrored_Storage_Params;
use  Storage_Params.Non_Volatile.Stable.Mirrored_Storage_Params;
with Storages.Non_Volatile;
use  Storages.Non_Volatile;

package Storages.Non_Volatile.Stable.Mirrored_Storages is

   type Mirrored_Storage_Type is new Stable_Storage_Type with private;
   type Mirrored_Storage_Ref is access all Mirrored_Storage_Type'Class;

   function Create (Params : Mirrored_Storage_Params_Type)
                    return Mirrored_Storage_Ref;
   --  Create the mirrored storage identified by Params and a
   --  Mirrored_Storage_Type object, associate them
   --  and hand back a reference.

   function Open (Params : in Mirrored_Storage_Params_Type;
                  Mode   : in Mode_Type := In_Mode)
                  return Mirrored_Storage_Ref;
   --  Open an already existing mirrored storage
   --  identified by Params, create a new
   --  Mirrored_Storage_Type object, associate them and
   --  hand back a reference.

   procedure Read (Storage : in out Mirrored_Storage_Type;
                   Item    : out Stream_Element_Array;
                   Last    : out Stream_Element_Offset);
   --  Read a data item from the mirrored storage.

   procedure Write (Storage : in out Mirrored_Storage_Type;
                    Item    : in Stream_Element_Array);
   --  Append 'Data' to the mirrored storage.

   procedure Open (Storage : in out Mirrored_Storage_Type;
                   Mode    : in Mode_Type := In_Mode);
   --  Open an already existing mirrored file storage.

   procedure Close (Storage : in out Mirrored_Storage_Type);
   --  Close a mirrored file storage.

   procedure Delete (Storage : in out Mirrored_Storage_Type);
   --  Delete a mirrored file storage.

   function Get_Current_Size (Storage : in Mirrored_Storage_Type)
                              return Stream_Element_Count;
   --  Returns the current size of the mirrored file storage.

private
   type Mirrored_Members_Type is new Limited_Controlled with record
      Params : Mirrored_Storage_Params_Type;
      MirrorFile1 : Non_Volatile_Storage_Ref;
      MirrorFile2 : Non_Volatile_Storage_Ref;
      LogFile : Non_Volatile_Storage_Ref;
   end record;

   subtype Duplicata is Integer range 1 .. 2;

   --  Encapsulating the data of the stable storage in a controlled type
   --  gives us the possibility to close the stream's files when they cease to
   --  exist. This is very useful to avoid dangling open files in case of
   --  corrupted streams, where 'Read will raise exceptions.

   procedure Cleanup (Storage : in out Mirrored_Storage_Type;
                     Original : in Duplicata);

   procedure Finalize (Members : in out Mirrored_Members_Type);

   type Mirrored_Storage_Type is new Stable_Storage_Type with
      record
         Members : Mirrored_Members_Type;
      end record;

end Storages.Non_Volatile.Stable.Mirrored_Storages;

