-------------------------------------------------------------------------------
--                                                                           --
--                            Non-Volatile Storages                          --
--                                                                           --
--  Data stored in non volatile storage is not lost when a program           --
--  terminates. In case of program crashes, data stored on non volatile      --
--  storage may get corrupted.                                               --
--                                                                           --
--  Since the data is not lost on program termination, the type provides     --
--  additional operations for housekeeping.                                  --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/01/05 12:57:08 $
--  $RCSfile: storages-non_volatile.ads,v $
--  $Revision: 1.1.1.1 $
--                                                                           --
-------------------------------------------------------------------------------

package Storages.Non_Volatile is

   type Non_Volatile_Storage_Type is abstract new Storage_Type with private;
   type Non_Volatile_Storage_Ref is access all Non_Volatile_Storage_Type'Class;

   --  function Create (Parameters : in ...) return Non_Volatile_Storage_Ref;
   --  Create a new storage object and associate it with the storage device.
   --  Such a procedure must be written for each concrete storage.
   --  The parameters may vary, depending on the storage device. One of the
   --  parameters should be an identifier that allows to uniquely identify
   --  the storage device.

   --  function Open (Parameters : in ...;
   --                 Mode : in Mode_Type) return Non_Volatile_Storage_Ref;
   --  Create a new non volatile storage object and associate it with
   --  an already existing storage identified by the parameters. Depending on
   --  the storage type, this function can be used with different modes.

   type Mode_Type is (In_Mode, Out_Mode, In_Out_Mode);

   procedure Open (Storage : in out Non_Volatile_Storage_Type;
                   Mode    : in Mode_Type := In_Mode) is abstract;
   --  Associate the storage object to the real storage device.
   --  This can also be seen as: establish a connection between the object
   --  and the real storage device.

   procedure Close (Storage : in out Non_Volatile_Storage_Type) is abstract;
   --  Deassociate the storage object from the real storage device.

   procedure Delete (Storage : in out Non_Volatile_Storage_Type) is abstract;
   --  Delete the real storage associated with the storage object.

   Storage_Not_Existant,
     Storage_Existant,
     Storage_Corrupted,
     Mode_Not_Supported : exception;

private

   type Non_Volatile_Storage_Type is abstract new Storage_Type
     with null record;

end Storages.Non_Volatile;

