-------------------------------------------------------------------------------
--                              Memory Storages                              --
--                                                                           --
--  This implementation of volatile storage uses the base memory of the      --
--  local machine to store the data. Memory is allocated dynamically when    --
--  needed.                                                                  --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/01/05 12:57:08 $
--  $RCSfile: storages-volatile-memory_storages.adb,v $
--  $Revision: 1.1.1.1 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;

package body Storages.Volatile.Memory_Storages is

   ---------------------------------------------------------------------------
   --  A stream is a linked list of nodes, each node can hold 'Size' elements
   --  and is filled in the range Data (RIdx .. WIdx - 1). If a node is full,
   --  a new one is created and added to the list.

   --  Reading always begins in node 'Stream.data.First' at index 'RIdx'. Once
   --  a whole node has been read, it is discarded. Writing always takes place
   --  in node 'Stream.Data.Last' at index 'WIdx'.

   --  It is possible for intermediary nodes to be filled only partially.
   --  Initially, a stream has no nodes. It also has no nodes at all if
   --  the whole stream has been read.

   type Node (Size : Stream_Element_Count) is
      record
         Data : Stream_Element_Array (1 .. Size);
         RIdx : Stream_Element_Offset := 1;
         WIdx : Stream_Element_Offset := 1;
         Next : Node_Ptr;
      end record;

   Min_Length : constant Stream_Element_Count := 256;
   --  Each node in the list is at least 'Min_Length' elements long.

   ---------------------------------------------------------------------------

   procedure Free is
     new Ada.Unchecked_Deallocation (Node, Node_Ptr);

   ---------------------------------------------------------------------------

   procedure Finalize
     (Data : in out Stream_Data)
      --  Deallocate any nodes that still are in the list.
   is
      Old : Node_Ptr;
   begin
      while Data.First /= null loop
         Old := Data.First; Data.First := Old.Next; Free (Old);
      end loop;
      Data.Last   := null;
      Data.Length := 0;
   end Finalize;

   ---------------------------------------------------------------------------

   procedure Read
     (Storage : in out Memory_Storage_Type;
      Item    :    out Stream_Element_Array;
      Last    :    out Stream_Element_Offset)
   is
      First   : Node_Ptr renames Storage.Data.First;
      Old     : Node_Ptr;
      Count   : Stream_Element_Count := Item'Length;
      Offset  : Stream_Element_Offset := Item'First;
      To_Read : Stream_Element_Count;
   begin
      while First /= null and Count > 0 loop
         To_Read :=
           Stream_Element_Count'Min (First.WIdx - First.RIdx, Count);
         Item (Offset .. Offset + To_Read - 1) :=
           First.Data (First.RIdx .. First.RIdx + To_Read - 1);
         Count      := Count - To_Read;
         Offset     := Offset + To_Read;
         First.RIdx := First.RIdx + To_Read;
         if First.RIdx >= First.WIdx then
            --  End of this node reached: deallocate.
            Old := First; First := First.Next; Free (Old);
         end if;
      end loop;
      Last := Offset - 1;
      Storage.Data.Length := Storage.Data.Length - (Offset - Item'First);
      if First = null then
         Storage.Data.Last := null;
      end if;
   end Read;

   ---------------------------------------------------------------------------

   procedure Write
     (Storage : in out Memory_Storage_Type;
      Item    : in     Stream_Element_Array)
   is
   begin
      if Storage.Data.Last = null then
         --  We need a new node...
         Storage.Data.Last :=
           new Node (Stream_Element_Count'Max (Item'Length, Min_Length));
         if Storage.Data.First = null then
            Storage.Data.First := Storage.Data.Last;
         end if;
      end if;
      if Storage.Data.Last.WIdx + Item'Length - 1 > Storage.Data.Last.Size then
         --  The data won't fit: simply put it in a new node!
         Storage.Data.Last.Next :=
           new Node (Stream_Element_Count'Max (Item'Length, Min_Length));
         Storage.Data.Last := Storage.Data.Last.Next;
      end if;
      Storage.Data.Last.Data (Storage.Data.Last.WIdx ..
                             Storage.Data.Last.WIdx + Item'Length - 1) := Item;
      Storage.Data.Last.WIdx := Storage.Data.Last.WIdx + Item'Length;
      Storage.Data.Length    := Storage.Data.Length + Item'Length;
   end Write;

   ---------------------------------------------------------------------------

   function To_Stream_Element_Array
     (Storage : in Memory_Storage_Type)
     return Stream_Element_Array
   is
      Res    : Stream_Element_Array (1 .. Storage.Data.Length);
      Ptr    : Node_Ptr := Storage.Data.First;
      Offset : Stream_Element_Offset := 1;
   begin
      while Ptr /= null loop
         Res (Offset .. Offset + Ptr.WIdx - Ptr.RIdx - 1) :=
           Ptr.Data (Ptr.RIdx .. Ptr.WIdx - 1);
         Offset := Offset + Ptr.WIdx - Ptr.RIdx;
         Ptr := Ptr.Next;
      end loop;
      return Res (1 .. Offset - 1); --  Should be the same as 'Res'.
   end To_Stream_Element_Array;

   ---------------------------------------------------------------------------

   procedure Clear
     (Storage : in out Memory_Storage_Type)
   is
   begin
      Finalize (Storage.Data);
   end Clear;

   ---------------------------------------------------------------------------

   function Get_Current_Size (Storage : in Memory_Storage_Type)
                              return Stream_Element_Count is
   begin
      return Storage.Data.Length;
   end Get_Current_Size;

end Storages.Volatile.Memory_Storages;


