-------------------------------------------------------------------------------
--                              Memory Storages                              --
--                                                                           --
--  This implementation of volatile storage uses the base memory of the      --
--  local machine to store the data. Memory is allocated dynamically when    --
--  needed.                                                                  --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/08/18 11:48:52 $
--  $RCSfile: storages-volatile-memory_storages.ads,v $
--  $Revision: 1.2 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Streams; use Ada.Streams;
with Ada.Finalization; use Ada.Finalization;

package Storages.Volatile.Memory_Storages is

   type Memory_Storage_Type is new Volatile_Storage_Type with private;
   type Memory_Storage_Ref is access all Memory_Storage_Type'Class;

   procedure Read (Storage : in out Memory_Storage_Type;
                   Item    : out Stream_Element_Array;
                   Last    : out Stream_Element_Offset);
   --  Read a data item from the storage

   procedure Write (Storage : in out Memory_Storage_Type;
                    Item    : in Stream_Element_Array);
   --  Append 'Data' to the storage

   function To_Stream_Element_Array (Storage : in Memory_Storage_Type)
      return Stream_Element_Array;

   procedure Clear (Storage : in out Memory_Storage_Type);
   --  Throw away all the data in the storage

   function Get_Current_Size (Storage : in Memory_Storage_Type)
      return Stream_Element_Count;

private

   type Node (<>);
   type Node_Ptr is access Node;

   type Stream_Data is new Limited_Controlled with
     record
        First, Last : Node_Ptr;
        Length      : Stream_Element_Count := 0;
     end record;

   --  Encapsulating the data of the memory storage in a controlled type
   --  gives us the possibility to free the stream's memory when it ceases to
   --  exist. This is very useful to avoid memory leaks in case of corrupted
   --  streams, where 'Read will raise exceptions.

   procedure Finalize
     (Data : in out Stream_Data);

   type Memory_Storage_Type is new Volatile_Storage_Type with
      record
         Data : Stream_Data;
      end record;

end Storages.Volatile.Memory_Storages;
