-------------------------------------------------------------------------------
--                             Unbounded Buffers                             --
--                                                                           --
--  The unbounded buffer uses a memory storage to store the data of calls    --
--  to Write. Only when Flush is called, the data is written to the          --
--  associated non-volatile storage.                                         --
--  On the first read, the *entire* data is read from the non-volatile       --
--  into memory. Successive reads will read from there.                      --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/01/27 15:23:48 $
--  $RCSfile: buffers-unbounded.adb,v $
--  $Revision: 1.1 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;

package body Buffers.Unbounded is

   procedure Free is new Ada.Unchecked_Deallocation
     (Memory_Storage_Type'Class, Memory_Storage_Ref);

   procedure Read (Buffer : in out Unbounded_Buffer_Type;
                   Item    : out Stream_Element_Array;
                   Last    : out Stream_Element_Offset) is
   begin
      if Buffer.Memory_Stream = null then
         declare
            S : Stream_Element_Array (1 .. Stream_Element_Offset
                                      (Get_Current_Size (Buffer.Storage.all)));
            Last : Stream_Element_Offset;
         begin
            Read (Buffer.Storage.all, S, Last);
            Buffer.Memory_Stream := new Memory_Storage_Type;
            Write (Buffer.Memory_Stream.all, S);
         end;
      end if;
      Read (Buffer.Memory_Stream.all, Item, Last);
   end Read;

   procedure Write (Buffer : in out Unbounded_Buffer_Type;
                    Item    : in Stream_Element_Array) is
   begin
      if Buffer.Memory_Stream = null then
         Buffer.Memory_Stream := new Memory_Storage_Type;
      end if;
      Write (Buffer.Memory_Stream.all, Item);
   end Write;

   procedure Flush (Buffer : in out Unbounded_Buffer_Type) is
   begin
      Write (Buffer.Storage.all,
             To_Stream_Element_Array (Buffer.Memory_Stream.all));
      Clear (Buffer.Memory_Stream.all);
      Free (Buffer.Memory_Stream);
   end Flush;

   procedure Finalize (Buffer : in out Unbounded_Buffer_Type) is
   begin
      Free (Buffer.Memory_Stream);
   end Finalize;

end Buffers.Unbounded;

