-------------------------------------------------------------------------------
--                                  Streams                                  --
--                                                                           --
--  The package streams presents the user with a concrete stream type.       --
--  (see ARM 13.13 for more detail). There are two version of streams,       --
--  normal streams, and buffered streams.                                    --
--                                                                           --
--  Normal streams have an associated storage, and they write the data to    --
--  that storage on every call to Write. When Read is called, the data is    --
--  read from the storage.                                                   --
--                                                                           --
--  Buffered streams are a little different. They write the data into an     --
--  intermediate (volatile) storage (a buffer), and only write the data to   --
--  the associated non-volatile storage under certain conditions (for        --
--  instance when the buffer is full.) To force the data to be written       --
--  to the non-volatile storage, Flush must be called.                       --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2000/01/31 14:20:10 $
--  $RCSfile: streams.ads,v $
--  $Revision: 1.2 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Streams; use Ada.Streams;
with Storages; use Storages;
with Buffers; use Buffers;

package Streams is

   ---------------------------
   --  Abstract_Stream_Type --
   ---------------------------

   type Abstract_Stream_Type is abstract new
     Ada.Streams.Root_Stream_Type with private;

   type Abstract_Stream_Ref is access all Abstract_Stream_Type'Class;

   ------------------
   --  Stream_Type --
   ------------------

   type Stream_Type (Storage : access Storage_Type'Class) is new
     Abstract_Stream_Type with private;

   type Stream_Ref is access all Stream_Type'Class;

   procedure Read (Stream : in out Stream_Type;
                   Item   : out Stream_Element_Array;
                   Last   : out Stream_Element_Offset);
   --  Read a data item from the stream. See [RM, 13.13.1(8)].

   procedure Write (Stream : in out Stream_Type;
                    Item   : in Stream_Element_Array);
   --  Append 'Data' to the stream. See [RM, 13.13.1(9)].

   ---------------------------
   --  Buffered_Stream_Type --
   ---------------------------

   type Buffered_Stream_Type (Buffer : access Buffer_Type'Class) is new
     Abstract_Stream_Type with private;

   type Buffered_Stream_Ref is access all Buffered_Stream_Type'Class;

   procedure Read (Stream : in out Buffered_Stream_Type;
                   Item   : out Stream_Element_Array;
                   Last   : out Stream_Element_Offset);
   --  Read a data item from the stream. See [RM, 13.13.1(8)].

   procedure Write (Stream : in out Buffered_Stream_Type;
                    Item   : in Stream_Element_Array);
   --  Append 'Data' to the stream. See [RM, 13.13.1(9)].

   procedure Flush (Stream : in out Buffered_Stream_Type);
   --  Write the data in the stream out to the associated non-volatile
   --  storage

private

   type Abstract_Stream_Type is abstract new
     Ada.Streams.Root_Stream_Type with null record;

   type Stream_Type (Storage : access Storage_Type'Class) is new
     Abstract_Stream_Type with null record;

   type Buffered_Stream_Type (Buffer : access Buffer_Type'Class) is new
     Abstract_Stream_Type with null record;

end Streams;
