-------------------------------------------------------------------------------
--             Transaction Identifiers - Task Attribute Interface            --
--                                                                           --
--  This package provides the bridge between the task attributes and the     --
--  task identifiers                                                         --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/23 11:19:13 $
--  $RCSfile: transaction_identifiers-task_attribute_interface.adb,v $
--  $Revision: 1.10 $
--                                                                           --
-------------------------------------------------------------------------------

with Recovery_Managers; use Recovery_Managers;
with Recovery_Support; use Recovery_Support;
with Transactional_Task_Attributes; use Transactional_Task_Attributes;
with Transactional_Tasks; use Transactional_Tasks;

with Linked_Lists_With_Refkey;
with Ada.Task_Identification; use Ada.Task_Identification;

package body Transaction_Identifiers.Task_Attribute_Interface is

   type String_Ref is access String;

   package Named_Transaction_Lists is new Linked_Lists_With_Refkey
     (Task_Id, String, String_Ref);

   function Get_Current_Transaction_Id return Transaction_Identifier_Type is
      Result : Transaction_Identifier_Ref :=
        Transactional_Tasks.Reference.Current_Tid;
   begin
      if Result = null then
         return Null_Transaction_Id;
      else
         return Transactional_Tasks.Reference.Current_Tid.all;
      end if;
   end Get_Current_Transaction_Id;

   function Get_Current_Transaction return Transaction_Ref is
      Result : Transaction_Identifier_Ref :=
        Transactional_Tasks.Reference.Current_Tid;
   begin
      if Result = null then
         return null;
      else
         return Transactional_Tasks.Reference.Current_Tid.Transaction;
      end if;
   end Get_Current_Transaction;

   procedure Begin_Transaction is
      Current_Transaction : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference);
   begin
      if Current_Transaction.Current_Tid = null or else
        Current_Transaction.Current_Tid.all = Null_Transaction_Id then
         Current_Transaction.Current_Tid :=
          new Transaction_Identifier_Type' (Get_New_Transaction_Id);
      else
         Current_Transaction.Current_Tid.all :=
            Get_Nested_Transaction_Id (Current_Transaction.Current_Tid.all);
      end if;
      Begin_Transaction (Get_Recovery_Manager.all,
                         Current_Transaction.Current_Tid.Transaction);
   end Begin_Transaction;

   procedure Join_Transaction (Transaction_Id : Transaction_Identifier_Type) is
      Current_Transaction : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference);
   begin
      if Current_Transaction.Current_Tid = null then
         --  Copy the tid
         Current_Transaction.Current_Tid :=
              new Transaction_Identifier_Type' (Transaction_Id);
         Join_Transaction (Current_Transaction.Current_Tid.Transaction.all);
      elsif Current_Transaction.Current_Tid.all = Null_Transaction_Id then
         Current_Transaction.Current_Tid.all := Transaction_Id;
         Join_Transaction (Current_Transaction.Current_Tid.Transaction.all);
      elsif Current_Transaction.Current_Tid.Transaction =
        Transaction_Id.Transaction then
         raise Already_Joined_Transaction;
      elsif Is_Ancestor_Or_Same
        (Current_Transaction.Current_Tid.Transaction.all,
         Transaction_Id.Transaction) then
         --  Copy the tid
         Current_Transaction.Current_Tid :=
              new Transaction_Identifier_Type' (Transaction_Id);
         Join_Transaction (Current_Transaction.Current_Tid.Transaction.all);
      else
         raise Already_Part_Of_Other_Transaction;
      end if;
   end Join_Transaction;

   procedure Join_Transaction (The_Task : Task_Id) is
      Current_Transaction : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference);
      Transaction_To_Join : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference
                                          (The_Task));
   begin
      if Current_Transaction.Current_Tid = null then
         --  Copy the tid
         Current_Transaction.Current_Tid :=
            new Transaction_Identifier_Type'
           (Transaction_To_Join.Current_Tid.all);
         Join_Transaction (Current_Transaction.Current_Tid.Transaction.all);
      elsif Current_Transaction.Current_Tid.all = Null_Transaction_Id then
         Current_Transaction.Current_Tid.all :=
           Transaction_To_Join.Current_Tid.all;
         Join_Transaction (Current_Transaction.Current_Tid.Transaction.all);
      elsif Current_Transaction.Current_Tid.Transaction =
        Transaction_To_Join.Current_Tid.Transaction then
         raise Already_Joined_Transaction;
      elsif Is_Ancestor_Or_Same
        (Current_Transaction.Current_Tid.Transaction.all,
         Transaction_To_Join.Current_Tid.Transaction) then
         --  Copy the tid
         Current_Transaction.Current_Tid :=
            new Transaction_Identifier_Type'
           (Transaction_To_Join.Current_Tid.all);
         Join_Transaction (Current_Transaction.Current_Tid.Transaction.all);
      else
         raise Already_Part_Of_Other_Transaction;
      end if;
   end Join_Transaction;

   The_List : Named_Transaction_Lists.List_Type;

   procedure Begin_Or_Join_Transaction (Name : String) is
      T : Task_Id;
   begin
      T := Named_Transaction_Lists.Find (The_List, Name);
      Join_Transaction (T);
   exception
      when Named_Transaction_Lists.No_Such_Key =>
         Begin_Transaction;
         Named_Transaction_Lists.Insert_Element (The_List, Name, Current_Task);
   end Begin_Or_Join_Transaction;

   procedure Close_Transaction is
      Current_Transaction : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference);
   begin
      if Current_Transaction.Current_Tid = null or else
        Current_Transaction.Current_Tid.all = Null_Transaction_Id then
         raise No_Current_Transaction;
      end if;
      Close_Transaction (Current_Transaction.Current_Tid.Transaction.all);
   end Close_Transaction;

   procedure Abort_Transaction is
      Current_Transaction : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference);
   begin
      if Current_Transaction.Current_Tid = null then
         raise No_Current_Transaction;
      else
         Abort_Transaction (Current_Transaction.Current_Tid.Transaction.all);
         Current_Transaction.Current_Tid.all := Get_Parent_Transaction_Id
           (Current_Transaction.Current_Tid.all);
      end if;
   end Abort_Transaction;

   procedure Commit_Transaction is
      Current_Transaction : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference);
   begin
      if Current_Transaction.Current_Tid = null or else
        Current_Transaction.Current_Tid.all = Null_Transaction_Id then
         raise No_Current_Transaction;
      else
         Commit_Transaction (Current_Transaction.Current_Tid.Transaction.all);
         Current_Transaction.Current_Tid.all := Get_Parent_Transaction_Id
           (Current_Transaction.Current_Tid.all);
      end if;
   end Commit_Transaction;

   function Create (Params   : Non_Volatile_Params_Type'Class;
                    Creation : Creation_Operation_Type'Class;
                    Loading  : Loading_Operation_Type'Class;
                    Saving   : Saving_Operation_Type'Class;
                    Deletion : Deletion_Operation_Type'Class)
      return Memory_Object_Ref is
      Object : Memory_Object_Ref;
      Current_Transaction : Transactional_Task_Attribute_Ref :=
        Transactional_Task_Attribute_Ref (Transactional_Tasks.Reference);
   begin
      if Current_Transaction.Current_Tid = null or else
        Current_Transaction.Current_Tid.all = Null_Transaction_Id then
         raise No_Current_Transaction;
      else
         Create (Get_Recovery_Manager.all,
                 Object, Current_Transaction.Current_Tid.Transaction,
                 Params, Creation, Loading, Saving, Deletion);
         return Object;
      end if;
   end Create;

end Transaction_Identifiers.Task_Attribute_Interface;





