-------------------------------------------------------------------------------
--                                Transactions                               --
--                                                                           --
--  This package provides transactions                                       --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/23 11:19:13 $
--  $RCSfile: transactions.adb,v $
--  $Revision: 1.9 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Debug; use Debug;

with Recovery_Managers; use Recovery_Managers;
with Recovery_Support; use Recovery_Support;

package body Transactions is

   --  We can have concurrent calls to Get_New_Transaction_Id
   --  We therefore need a protected distributor

   protected Distributor is

      procedure Get_Next_Simple_Transaction
         (Simple : out Simple_Transaction_Identifier_Type);

   private

      Next_Free_Number : Positive := 1;

      --  This must be changed if we include distribution
      Node_Number : Positive := 123;

   end Distributor;

   protected body Distributor is

      procedure Get_Next_Simple_Transaction
         (Simple : out Simple_Transaction_Identifier_Type) is
      begin
         Simple.Number := Next_Free_Number;
         Next_Free_Number := Next_Free_Number + 1;
         Simple.Node.Number := Node_Number;
      end Get_Next_Simple_Transaction;

   end Distributor;

   procedure Check_And_Then_Add (Transaction : in out Transaction_Type;
                                 Object : in Memory_Object_Ref) is
   begin
      Transaction.Protection.Check_And_Then_Add (Object);
   end Check_And_Then_Add;

   procedure Check_And_Then_Add_Modified
     (Transaction : in out Transaction_Type;
      Object : in Memory_Object_Ref) is
   begin
      Transaction.Protection.Check_And_Then_Add_Modified (Object);
   end Check_And_Then_Add_Modified;

   procedure Decrease_Reference_Count (T : in Transaction_Ref) is
   begin
      T.Protection.Decrement_Reference_Count;
   end Decrease_Reference_Count;

   procedure Increase_Reference_Count (T : in Transaction_Ref) is
   begin
      T.Protection.Increment_Reference_Count;
   end Increase_Reference_Count;

   procedure Free is new Ada.Unchecked_Deallocation
     (Transaction_Type'Class, Transaction_Ref);

   procedure Check_Reference_Count (T : in out Transaction_Ref) is
   begin
      if T.Protection.Get_Reference_Count = 0 then
         Put_Line ("Deleted transaction");
         Free (T);
      end if;
   end Check_Reference_Count;

   protected body Concurrency_Protection_Type is

      procedure Decrement_Participants is
      begin
         Number_Of_Participants := Number_Of_Participants - 1;
      end Decrement_Participants;

      procedure Increment_Participants is
      begin
         Number_Of_Participants := Number_Of_Participants + 1;
      end Increment_Participants;

      procedure Set_Number_Of_Participants (Number : in Natural) is
      begin
         Number_Of_Participants := Number;
      end Set_Number_Of_Participants;

      function Get_Number_Of_Participants return Natural is
      begin
         return Number_Of_Participants;
      end Get_Number_Of_Participants;

      procedure Decrement_Reference_Count is
      begin
         Reference_Count := Reference_Count - 1;
--          if Children /= null then
--             Put_Line ("Topref :" & Integer'Image (Reference_Count));
--          else
--             Put_Line ("Ref :" & Integer'Image (Reference_Count));
--          end if;
      end Decrement_Reference_Count;

      procedure Increment_Reference_Count is
      begin
         Reference_Count := Reference_Count + 1;
--          if Children /= null then
--             Put_Line ("Topref :" & Integer'Image (Reference_Count));
--          else
--             Put_Line ("Ref :" & Integer'Image (Reference_Count));
--          end if;
      end Increment_Reference_Count;

      procedure Set_Reference_Count (Number : in Natural) is
      begin
         Reference_Count := Number;
      end Set_Reference_Count;

      function Get_Reference_Count return Natural is
      begin
         return Reference_Count;
      end Get_Reference_Count;

      procedure Set_Transaction_Status (Status : in Transaction_Status_Type) is
      begin
         Transaction_Status := Status;
      end Set_Transaction_Status;

      function Get_Transaction_Status return Transaction_Status_Type is
      begin
         return Transaction_Status;
      end Get_Transaction_Status;

      entry Suspend_Task when Transaction_Status = Aborted or
                              Transaction_Status = Committed is
      begin
         null;
      end Suspend_Task;

      procedure Add_Child (Child : Transaction_Ref) is
      begin
         Child_Lists.Insert_Element (Child_List, Child);
         Increment_Reference_Count;
      end Add_Child;

      procedure Remove_Child (Child : Transaction_Ref) is
      begin
         Child_Lists.Delete_Element (Child_List, Child);
         Decrement_Reference_Count;
      end Remove_Child;

      procedure Check_And_Then_Add (Object : in Memory_Object_Ref) is
      begin
         Object_Lists.Insert_If_Not_Already (Accessed_Object_List, Object);
      end Check_And_Then_Add;

      procedure Check_And_Then_Add_Modified (Object : in Memory_Object_Ref) is
      begin
         Object_Lists.Insert_If_Not_Already (Accessed_Object_List, Object);
         Object_Lists.Insert_If_Not_Already (Modified_Object_List, Object);
      end Check_And_Then_Add_Modified;

      procedure Append_Objects (Objects : Object_Lists.List_Type) is
      begin
         Object_Lists.Merge_Lists (Accessed_Object_List, Objects);
      end Append_Objects;

      procedure Append_Objects_Modified (Objects : Object_Lists.List_Type) is
      begin
         Object_Lists.Merge_Lists (Modified_Object_List, Objects);
      end Append_Objects_Modified;

      procedure Delete_Objects is
      begin
         Object_Lists.Delete_List (Accessed_Object_List);
      end Delete_Objects;

      procedure Delete_Objects_Modified is
      begin
         Object_Lists.Delete_List (Modified_Object_List);
      end Delete_Objects_Modified;

      function Get_Object_List return Object_Lists.List_Type is
      begin
         return Accessed_Object_List;
      end Get_Object_List;

      function Get_Object_List return Accessed_Objects_Type is
         N : Natural := Object_Lists.Count_Elements (Accessed_Object_List);
         Result : Accessed_Objects_Type (1 .. N);
         Tmp : Object_Lists.List_Type := Accessed_Object_List;
      begin
         for I in Result'Range loop
            Result (I) := Tmp.Element;
            Tmp := Tmp.Next;
         end loop;
         return Result;
      end Get_Object_List;

      function Get_Object_List_Modified return Object_Lists.List_Type is
      begin
         return Modified_Object_List;
      end Get_Object_List_Modified;

      function Get_Object_List_Modified return Accessed_Objects_Type is
         N : Natural := Object_Lists.Count_Elements (Modified_Object_List);
         Result : Accessed_Objects_Type (1 .. N);
         Tmp : Object_Lists.List_Type := Modified_Object_List;
      begin
         for I in Result'Range loop
            Result (I) := Tmp.Element;
            Tmp := Tmp.Next;
         end loop;
         return Result;
      end Get_Object_List_Modified;

   end Concurrency_Protection_Type;

   function Create_New_Toplevel_Transaction return Transaction_Ref is
      Temp_Ref : Transaction_Ref := new Transaction_Type;
   begin
      Temp_Ref.Protection.Set_Transaction_Status (Open);
      Temp_Ref.Protection.Set_Reference_Count (1);
      Temp_Ref.Protection.Set_Number_Of_Participants (1);
      Distributor.Get_Next_Simple_Transaction (Temp_Ref.Identifier);
      return Temp_Ref;
   end Create_New_Toplevel_Transaction;

   procedure Create_Nested_Transaction
     (Parent : in out Transaction_Type;
      New_Transaction : out Transaction_Ref) is
      Temp_Ref : Transaction_Ref := new Transaction_Type;
   begin
      Temp_Ref.Parent := Self (Parent);
      Parent.Protection.Add_Child (Temp_Ref);
      Temp_Ref.Protection.Set_Transaction_Status (Open);
      Temp_Ref.Protection.Set_Reference_Count (1);
      Temp_Ref.Protection.Set_Number_Of_Participants (1);
      Distributor.Get_Next_Simple_Transaction (Temp_Ref.Identifier);
      New_Transaction := Temp_Ref;
   end Create_Nested_Transaction;

   procedure Join_Transaction (Transaction : in out Transaction_Type) is
   begin
      case Transaction.Protection.Get_Transaction_Status is
         when Open =>
            Transaction.Protection.Increment_Participants;
         when Aborted =>
            raise Transaction_Aborted;
         when Committed =>
            raise Transaction_Already_Committed;
         when Closed =>
            raise Transaction_Closed;
      end case;
   end Join_Transaction;

   procedure Close_Transaction (Transaction : in out Transaction_Type) is
   begin
      case Transaction.Protection.Get_Transaction_Status is
         when Open =>
            Transaction.Protection.Set_Transaction_Status (Closed);
         when Aborted =>
            raise Transaction_Aborted;
         when Committed =>
            raise Transaction_Already_Committed;
         when Closed =>
            null;
      end case;
   end Close_Transaction;

   procedure Commit_Transaction (Transaction : in out Transaction_Type) is
   begin
      Transaction.Protection.Decrement_Participants;
      if Transaction.Protection.Get_Transaction_Status = Aborted then
         raise Transaction_Aborted;
      end if;
      if Transaction.Protection.Get_Number_Of_Participants = 0 then
         Commit_Transaction (Get_Recovery_Manager.all, Self (Transaction));
         Transaction.Protection.Set_Transaction_Status (Committed);
         if Transaction.Parent /= null then
            --  pass accessed objects to parent
            Transaction.Parent.Protection.Append_Objects
              (Transaction.Protection.Get_Object_List);
            Transaction.Protection.Delete_Objects;
            Transaction.Parent.Protection.Append_Objects_Modified
              (Transaction.Protection.Get_Object_List_Modified);
            Transaction.Protection.Delete_Objects_Modified;
            Transaction.Parent.Protection.Remove_Child (Self (Transaction));
         else
            --  delete accessed objects
            Transaction.Protection.Delete_Objects;
            Transaction.Protection.Delete_Objects_Modified;
         end if;
      else
         Transaction.Protection.Suspend_Task;
         if Transaction.Protection.Get_Transaction_Status = Aborted then
            raise Transaction_Aborted;
         end if;
      end if;
   end Commit_Transaction;

   procedure Abort_Transaction (Transaction : in out Transaction_Type) is
   begin
      Transaction.Protection.Decrement_Participants;
      if Transaction.Protection.Get_Transaction_Status /= Aborted then
         Abort_Transaction (Get_Recovery_Manager.all, Self (Transaction));
         Transaction.Protection.Set_Transaction_Status (Aborted);
         --  delete accessed objects
         Transaction.Protection.Delete_Objects;
         Transaction.Protection.Delete_Objects_Modified;
         if Transaction.Parent /= null then
            Transaction.Parent.Protection.Remove_Child (Self (Transaction));
         end if;
      end if;
   end Abort_Transaction;

   procedure Set_Final_Transaction_Status
     (Transaction : in out Transaction_Type;
      Status : in Transaction_Status_Type) is
   begin
      Transaction.Protection.Set_Transaction_Status (Status);
   end Set_Final_Transaction_Status;

   function Self (Transaction : in Transaction_Type) return
     Transaction_Ref is
   begin
      return Transaction.Self.Obj.all'Unchecked_Access;
   end Self;

   procedure Finalize (Transaction : in out Transaction_Type) is
   begin
      Transaction.Protection.Delete_Objects;
   end Finalize;

   function Is_Ancestor_Or_Same (Transaction : in Transaction_Type;
                                 Parent : in Transaction_Ref) return Boolean is
      Tmp : Transaction_Ref := Self (Transaction);
   begin
      while Tmp /= null loop
         if Tmp = Parent then
            return True;
         end if;
         Tmp := Tmp.Parent;
      end loop;
      return False;
   end Is_Ancestor_Or_Same;

   function Is_Toplevel (Transaction : Transaction_Type) return Boolean is
   begin
      return Transaction.Parent = null;
   end Is_Toplevel;

   function Get_Parent (Transaction : Transaction_Type)
      return Transaction_Ref is
   begin
      return Transaction.Parent;
   end Get_Parent;

   function Get_Simple_Transaction_Identifier (Transaction : Transaction_Type)
      return Simple_Transaction_Identifier_Type is
   begin
      return Transaction.Identifier;
   end Get_Simple_Transaction_Identifier;

   function Transaction_Text (Transaction : Transaction_Type)
      return String is
   begin
      return Natural'Image (Transaction.Identifier.Number);
   end Transaction_Text;

   function Transaction_Text (Transaction : Simple_Transaction_Identifier_Type)
      return String is
   begin
      return Natural'Image (Transaction.Number);
   end Transaction_Text;

   function Get_Accessed_Objects (Transaction : Transaction_Type)
      return Accessed_Objects_Type is
   begin
      return Transaction.Protection.Get_Object_List;
   end Get_Accessed_Objects;

   function Get_Modified_Objects (Transaction : Transaction_Type)
      return Accessed_Objects_Type is
   begin
      return Transaction.Protection.Get_Object_List_Modified;
   end Get_Modified_Objects;

end Transactions;
