-------------------------------------------------------------------------------
--                                Transactions                               --
--                                                                           --
--  This package provides transactions                                       --
--                                                                           --
--  $Author: jkienzle $
--  $Date: 2001/04/23 11:19:14 $
--  $RCSfile: transactions.ads,v $
--  $Revision: 1.9 $
--                                                                           --
-------------------------------------------------------------------------------

with Memory_Objects; use Memory_Objects;
with Ada.Finalization; use Ada.Finalization;

with Linked_Lists;

package Transactions is

   type Transaction_Type is tagged limited private;

   type Transaction_Ref is access all Transaction_Type'Class;

   function Create_New_Toplevel_Transaction return Transaction_Ref;

   procedure Create_Nested_Transaction
     (Parent : in out Transaction_Type;
      New_Transaction : out Transaction_Ref);

   procedure Decrease_Reference_Count (T : in Transaction_Ref);

   procedure Increase_Reference_Count (T : in Transaction_Ref);

   procedure Check_Reference_Count (T : in out Transaction_Ref);

   procedure Check_And_Then_Add (Transaction : in out Transaction_Type;
                                 Object : in Memory_Object_Ref);

   procedure Check_And_Then_Add_Modified
     (Transaction : in out Transaction_Type;
      Object : in Memory_Object_Ref);

   procedure Close_Transaction (Transaction : in out Transaction_Type);

   procedure Abort_Transaction (Transaction : in out Transaction_Type);

   procedure Commit_Transaction (Transaction : in out Transaction_Type);

   procedure Join_Transaction (Transaction : in out Transaction_Type);

   No_Current_Transaction,
     Already_Joined_Transaction,
     Already_Part_Of_Other_Transaction,
     Transaction_Already_Committed,
     Transaction_Aborted,
     Transaction_Closed : exception;

   --  The following procedures are called by the recovery manager,
   --  once the outcome of a transaction has been determined

   type Transaction_Status_Type is
      (Open, Closed, Committed, Aborted);

   procedure Set_Final_Transaction_Status
     (Transaction : in out Transaction_Type;
      Status : in Transaction_Status_Type);

   function Is_Ancestor_Or_Same (Transaction : in Transaction_Type;
                                 Parent : in Transaction_Ref) return Boolean;

   function Is_Toplevel (Transaction : Transaction_Type) return Boolean;

   function Get_Parent (Transaction : Transaction_Type) return Transaction_Ref;

   type Simple_Transaction_Identifier_Type is private;

   Null_Transaction_Identifier : constant Simple_Transaction_Identifier_Type;

   function Get_Simple_Transaction_Identifier (Transaction : Transaction_Type)
      return Simple_Transaction_Identifier_Type;

   function Transaction_Text (Transaction : Transaction_Type)
      return String;

   function Transaction_Text (Transaction : Simple_Transaction_Identifier_Type)
      return String;

   type Accessed_Objects_Type is array (Natural range <>) of
     Memory_Object_Ref;

   function Get_Accessed_Objects (Transaction : Transaction_Type)
     return Accessed_Objects_Type;

   function Get_Modified_Objects (Transaction : Transaction_Type)
     return Accessed_Objects_Type;

private

   type Processing_Node_Type is record
      Number : Natural := 1;
   end record;

   type Simple_Transaction_Identifier_Type is record
      Number : Natural;
      Node   : Processing_Node_Type;
   end record;

   Null_Transaction_Identifier : constant
     Simple_Transaction_Identifier_Type := (Number => 0,
                                            Node => (Number => 1));

   type Selfer (Obj : access Transaction_Type'Class) is limited null record;

   package Object_Lists is new Linked_Lists (Memory_Object_Ref);

   package Child_Lists is new Linked_Lists (Transaction_Ref);

   protected type Concurrency_Protection_Type is

      procedure Decrement_Participants;

      procedure Increment_Participants;

      procedure Set_Number_Of_Participants (Number : in Natural);

      function Get_Number_Of_Participants return Natural;

      procedure Decrement_Reference_Count;

      procedure Increment_Reference_Count;

      procedure Set_Reference_Count (Number : in Natural);

      function Get_Reference_Count return Natural;

      procedure Set_Transaction_Status (Status : in Transaction_Status_Type);

      function Get_Transaction_Status return Transaction_Status_Type;

      entry Suspend_Task;

      procedure Add_Child (Child : Transaction_Ref);

      procedure Remove_Child (Child : Transaction_Ref);

      procedure Check_And_Then_Add (Object : in Memory_Object_Ref);

      procedure Append_Objects (Objects : Object_Lists.List_Type);

      procedure Delete_Objects;

      function Get_Object_List return Object_Lists.List_Type;

      function Get_Object_List return Accessed_Objects_Type;

      procedure Check_And_Then_Add_Modified (Object : in Memory_Object_Ref);

      procedure Append_Objects_Modified (Objects : Object_Lists.List_Type);

      procedure Delete_Objects_Modified;

      function Get_Object_List_Modified return Object_Lists.List_Type;

      function Get_Object_List_Modified return Accessed_Objects_Type;

   private

      Number_Of_Participants : Natural := 0;
      Reference_Count : Natural := 0;
      Transaction_Status : Transaction_Status_Type := Open;
      Child_List : Child_Lists.List_Type;
      Accessed_Object_List : Object_Lists.List_Type;
      Modified_Object_List : Object_Lists.List_Type;

   end Concurrency_Protection_Type;

   type Transaction_Type is new Limited_Controlled with record
      Self  : Selfer (Transaction_Type'Access);
      Protection : Concurrency_Protection_Type;
      Parent : Transaction_Ref := null;
      Level : Natural := 0;
      Identifier : Simple_Transaction_Identifier_Type;
   end record;

   function Self (Transaction : in Transaction_Type) return
     Transaction_Ref;

   procedure Finalize (Transaction : in out Transaction_Type);

end Transactions;
