-------------------------------------------------------------------------------
--                               Keyed Lists                                 --
--                                                                           --
--  This is a utility package to create keyed lists, mainly used for         --
--  transaction lists...                                                     --
--                                                                           --
--  $Author: xcaron $
--  $Date: 2000/11/30 12:26:18 $
--  $RCSfile: keyed_lists.adb,v $
--  $Revision: 1.6 $
--                                                                           --
-------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Debug; use Debug;

package body Keyed_Lists is

   procedure Free is new Ada.Unchecked_Deallocation
     (Element_Node_Type, Element_Node_Ref);

   procedure Free is new Ada.Unchecked_Deallocation
     (Key_Node_Type, Key_Node_Ref);

   function Find (List : List_Type;
                  Key : Key_Type) return Element_Node_Ref is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            return Tmp.Elements;
         end if;
         Tmp := Tmp.Next;
      end loop;
      raise No_Such_Key;
   end Find;

   procedure Replace_Key (List : in out List_Type;
                          Key : in Key_Type;
                          New_Key : in Key_Type) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            Tmp.Key := New_Key;
            return;
         end if;
         Tmp := Tmp.Next;
      end loop;
      raise No_Such_Key;
   end Replace_Key;

   procedure Create_Key (List : in out List_Type;
                         Key : in Key_Type;
                         Info : in Additional_Info_Type) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            raise Key_Exists;
         end if;
         Tmp := Tmp.Next;
      end loop;
      Tmp := new Key_Node_Type;
      Tmp.Key := Key;
      Tmp.Next := List.First;
      List.First := Tmp;
   end Create_Key;

   procedure Set_Info (List : in out List_Type;
                       Key : in Key_Type;
                       Info : in Additional_Info_Type) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            Tmp.Info := Info;
            return;
         end if;
         Tmp := Tmp.Next;
      end loop;
      raise No_Such_Key;
   end Set_Info;

   function Get_Info (List : List_Type;
                      Key : Key_Type) return Additional_Info_Type is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            return Tmp.Info;
         end if;
         Tmp := Tmp.Next;
      end loop;
      raise No_Such_Key;
   end Get_Info;

   procedure Insert (List : in out List_Type;
                     Key : in Key_Type;
                     Elements : in Element_Node_Ref) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            raise Key_Exists;
         end if;
      end loop;
      Tmp := new Key_Node_Type;
      Tmp.Next := List.First;
      List.First := Tmp;
      Tmp.Key := Key;
      Tmp.Elements := Elements;
   end Insert;

   procedure Merge_Lists (List : in out List_Type;
                          Dest_Key : in Key_Type;
                          Source_Key : in Key_Type) is
      Prev_Source_Tmp : Key_Node_Ref := List.First;
      Source_Tmp : Key_Node_Ref := List.First;
      Dest_Tmp : Key_Node_Ref := List.First;
   begin
      if Source_Key = Dest_Key then
         raise Program_Error;
      end if;
      while Source_Tmp /= null loop
         if Source_Tmp.Key = Source_Key then
            exit;
         end if;
         Prev_Source_Tmp := Source_Tmp;
         Source_Tmp := Source_Tmp.Next;
      end loop;
      if Source_Tmp = null then
         raise No_Such_Key;
      end if;
      while Dest_Tmp /= null loop
         if Dest_Tmp.Key = Dest_Key then
            exit;
         end if;
         Dest_Tmp := Dest_Tmp.Next;
      end loop;
      if Dest_Tmp /= null then
         declare
            Tmp2 : Element_Node_Ref := Source_Tmp.Elements;
         begin
            if Tmp2 /= null then
               while Tmp2.Next /= null loop
                  Tmp2 := Tmp2.Next;
               end loop;
               Tmp2.Next := Dest_Tmp.Elements;
               Dest_Tmp.Elements := Source_Tmp.Elements;
            end if;
         end;
         if Source_Tmp = List.First then
            List.First := List.First.Next;
         else
            Prev_Source_Tmp.Next := Source_Tmp.Next;
         end if;
         Free (Source_Tmp);
      else
         Source_Tmp.Key := Dest_Key;
      end if;

   end Merge_Lists;

   procedure Delete (List : in out List_Type;
                     Key : in Key_Type) is
      Tmp : Key_Node_Ref := List.First;
      Tmp2 : Key_Node_Ref := null;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            if Tmp2 = null then
               List.First := Tmp.Next;
            else
               Tmp2.Next := Tmp.Next;
            end if;
            --  now delete all the elements
            declare
               Tmp3 : Element_Node_Ref := Tmp.Elements;
               To_Free : Element_Node_Ref;
            begin
               while Tmp3 /= null loop
                  To_Free := Tmp3;
                  Tmp3 := Tmp3.Next;
                  Free (To_Free);
               end loop;
            end;
            Free (Tmp);
            return;
         end if;
         Tmp2 := Tmp;
         Tmp := Tmp.Next;
      end loop;
      raise No_Such_Key;
   end Delete;

   procedure Insert_Element (List : in out List_Type;
                             Key : in Key_Type;
                             Element : in Element_Type;
                             Create_Key : in Boolean := False) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            declare
               Tmp2 : Element_Node_Ref := new Element_Node_Type;
            begin
               Tmp2.Next := Tmp.Elements;
               Tmp.Elements := Tmp2;
               Tmp2.Element := Element;
               Put_Line ("List: Added element");
            end;
            return;
         end if;
         Tmp := Tmp.Next;
      end loop;
      --  The key does not yet exist
      if Create_Key then
         Tmp := new Key_Node_Type;
         Tmp.Next := List.First;
         Tmp.Key := Key;
         Tmp.Elements := new Element_Node_Type;
         Tmp.Elements.Element := Element;
         List.First := Tmp;
         Put_Line ("List: Created key and added element");
      else
         raise No_Such_Key;
      end if;
   end Insert_Element;

   procedure Append_Element (List : in out List_Type;
                             Key : in Key_Type;
                             Element : in Element_Type) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            declare
               Tmp2 : Element_Node_Ref := new Element_Node_Type;
               Tmp3 : Element_Node_Ref := Tmp.Elements;
            begin
               if Tmp3 = null then
                  Tmp.Elements := Tmp2;
                  Tmp2.Element := Element;
               else
                  while Tmp3.Next /= null loop
                     Tmp3.Next := Tmp2;
                     Tmp2.Element := Element;
                  end loop;
               end if;
            end;
            return;
         end if;
         Tmp := Tmp.Next;
      end loop;
      raise No_Such_Key;
   end Append_Element;

   procedure Delete_Element (List : in out List_Type;
                             Key : in Key_Type;
                             Element : in Element_Type) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         if Tmp.Key = Key then
            declare
               Tmp2 : Element_Node_Ref := Tmp.Elements;
               Tmp3 : Element_Node_Ref := Tmp.Elements;
            begin
               while Tmp2 /= null loop
                  if Tmp2.Element = Element then
                     if Tmp3 = Tmp2 then
                        Tmp.Elements := Tmp2.Next;
                        Free (Tmp2);
                     else
                        Tmp3.Next := Tmp2.Next;
                        Free (Tmp2);
                     end if;
                     return;
                  end if;
                  Tmp3 := Tmp2;
                  Tmp2 := Tmp2.Next;
               end loop;
               raise No_Such_Element;
            end;
         end if;
         Tmp := Tmp.Next;
      end loop;
      raise No_Such_Key;
   end Delete_Element;

   procedure Iterate_Through_Keys (List : in out List_Type) is
      Tmp : Key_Node_Ref := List.First;
   begin
      while Tmp /= null loop
         Action (Tmp.Key);
         Tmp := Tmp.Next;
      end loop;
   end Iterate_Through_Keys;

   procedure Finalize (List : in out List_Type) is
--       Tmp : Key_Node_Ref := List.First;
--       To_Free : Key_Node_Ref;
--       Tmp2 : Element_Node_Ref;
--       To_Free2 : Element_Node_Ref;
   begin
--       while Tmp /= null loop
--          Tmp2 := Tmp.Elements;
--          while Tmp2 /= null loop
--             To_Free2 := Tmp2;
--             Tmp2 := Tmp2.Next;
--             Free (To_Free2);
--          end loop;
--          To_Free := Tmp;
--          Tmp := Tmp.Next;
--          Free (To_Free);
--       end loop;
      null;
   end Finalize;

end Keyed_Lists;
